import numpy as np
import numpy.linalg as lg
import scipy.linalg as slg
import scipy.linalg.lapack as slapack
import scipy.sparse.linalg as splg
from math import *
import matplotlib.pyplot as plt
from mpl_toolkits import mplot3d
import scipy.sparse as sps
import scipy.interpolate as sinterp
import time as tm

import fdyn
import toolkit
from utils import *

from paneller import *
from body import *
from wing import *

class aircraft: #class to ease certain case studies for full aircraft
    #generate aircraft BEFORE patchcompose functions
    def parameter_report(self):
        print('Freestream parameters:')
        print('%10s %10s %10s %10s %10s %10s %10s' % ('a', 'b', 'p', 'q', 'r', 'Uinf', 'M'))
        print('%10f %10f %10f %10f %10f %10f %10f' % (degrees(self.a), degrees(self.b), self.p, self.q, self.r, self.Uinf, self.M))
        print('Environment parameters:')
        print('%10s %10s %10s %10s' % ('rho', 'mu', 'g', 'gamma'))
        print('%10f %10f %10f %10f' % (self.rho, self.mu, self.g, self.gamma))
        print('Geometry parameters:')
        print('%10s %10s %10s %10s' % ('Sref', 'cref', 'bref', 'AR'))
        print('%10f %10f %10f %10f' % (self.Sref, self.cref, self.bref, self.AR))
        if self.hascontrol():
            print('Control parameters:')
            controlnames_str=''
            controlvals_str=''
            for cname in self.controlset:
                controlnames_str+='%10s' % cname
                controlvals_str+='%10f' % degrees(self.controlset[cname].state)
            print(controlnames_str)
            print(controlvals_str)
        if not self.massavailable:
            print('Mass parameters not yet present')
        else:
            print('Mass parameters:')
            print('CG=(%10f %10f %10f)' % (self.CG[0], self.CG[1], self.CG[2]))
            print('%10s %10s %10s %10s %10s %10s %10s' % ('m', 'Ixx', 'Iyy', 'Izz', 'Ixy', 'Ixz', 'Iyz'))
            print('%10f %10f %10f %10f %10f %10f %10f' % (self.m, self.Inertia[0, 0], self.Inertia[1, 1], self.Inertia[2, 2], \
                self.Inertia[0, 1], self.Inertia[0, 2], self.Inertia[1, 2]))
    def addmass(self, m=0.0, Ixx=0.0, Iyy=0.0, Izz=0.0, Ixy=0.0, Ixz=0.0, Iyz=0.0, echo=True): #sets mass variables for the aircraft
        self.m=m
        self.Inertia=np.array([[Ixx, Ixy, Ixz], [Ixy, Iyy, Iyz], [Ixz, Iyz, Izz]])
        self.massavailable=True
        if echo:
            self.parameter_report()
    def dynamic_simulation(self, nstep=2e5, dt=1e-4, start_point=np.array([0.0, 0.0, 0.0]), \
        perturbations={}, trim=True, visc=True, onboard=np.zeros(3)):
        if not self.massavailable:
            print('WARNING: dynamic analysis requested without any mass data. Returning empty arrays')
            return np.array([]), np.array([]), np.array([]), np.array([])
        keys=['phi', 'theta', 'psi', 'u', 'v', 'w', 'p', 'q', 'r']
        for par in keys:
            if not par in perturbations:
                perturbations[par]=0.0
        perturbations=np.array([start_point[0], start_point[1], start_point[2], \
            perturbations['phi'], perturbations['theta'], perturbations['psi'], \
                perturbations['u'], perturbations['v'], perturbations['w'], \
                    perturbations['p'], perturbations['q'], perturbations['r']])
        coeffs=np.zeros(6)
        derivs=np.zeros((5, 6))
        if not trim:
            coeffs=np.array([-self.CX, self.CY, -self.CZ, -self.Cl, self.Cm, -self.Cn]) #note that we're using stability coordinates here
            if visc:
                coeffs+=np.array([-self.dCX, self.dCY, -self.dCZ, -self.dCl, self.dCm, -self.dCn])
        i=0
        for par in ['a', 'b', 'p', 'q', 'r']:
            derivs[i, :]=np.array([-self.stabderivative_dict[par]['dCX'], self.stabderivative_dict[par]['dCY'], -self.stabderivative_dict[par]['dCZ'], \
                -self.stabderivative_dict[par]['dCl'], self.stabderivative_dict[par]['dCm'], -self.stabderivative_dict[par]['dCn']])
            i+=1
        external_history, alpha_history, beta_history, euler_history, time_history=fdyn.tstep_solve(nstep, dt, self.rho, self.Uinf, self.g, \
            self.Sref, self.cref, self.bref, perturbations, onboard, self.Inertia, self.m, \
                coeffs, derivs)
        return external_history, alpha_history, beta_history, euler_history, time_history
    def balance(self, SM=0.1, echo=True): #balancing the aircraft for given static margin
        if not self.stabavailable:
            print('WARNING: stability derivatives not available for balancing. Performing in place')
            self.calcstab(echo=echo)
        loc_SM=-self.stabderivative_dict['a']['dCm']/self.stabderivative_dict['a']['dCL']
        self.transp_byvec(-(SM-loc_SM)*np.array([1.0, 0.0, 0.0])*self.cref)
        new_SM=-self.stabderivative_dict['a']['dCm']/self.stabderivative_dict['a']['dCL']
        if echo:
            print('==========Balancing========')
            print('%10s %10s' % ('SM init', 'SM final'))
            print('%10f %10f' % (loc_SM, new_SM))
            self.forces_report()
            if self.stabavailable:
                self.stabreport()
            print('===========================')
    def transp_byvec(self, vec): #transport pre-calculated forces by a 'vec' deallocation in CG
        mom_transp=np.cross(vec, np.array([self.CX, self.CY, self.CZ]))
        self.Cl-=mom_transp[0]/self.bref
        self.Cm-=mom_transp[1]/self.cref
        self.Cn-=mom_transp[2]/self.bref
        if self.stabavailable:
            pars=['a', 'b', 'p', 'q', 'r']
            for par in pars:
                dmom_transp=np.cross(vec, np.array([self.stabderivative_dict[par]['dCX'], self.stabderivative_dict[par]['dCY'], \
                    self.stabderivative_dict[par]['dCZ']]))
                self.stabderivative_dict[par]['dCl']-=dmom_transp[0]/self.bref
                self.stabderivative_dict[par]['dCm']-=dmom_transp[1]/self.cref
                self.stabderivative_dict[par]['dCn']-=dmom_transp[2]/self.bref
        if self.hascorrections:
            mom_transp=np.cross(vec, np.array([self.dCX, self.dCY, self.dCZ]))
            self.dCl-=mom_transp[0]/self.bref
            self.dCm-=mom_transp[1]/self.cref
            self.dCn-=mom_transp[2]/self.bref
    def transp_to_cg(self, CG):
        self.transp_byvec(-self.CG)
        self.transp_byvec(CG)
        self.CG=CG
    def __init__(self, sld, elems=[], Sref=0.0, cref=0.0, bref=0.0, echo=True, CG=np.array([0.0, 0.0, 0.0])):
        self.wings=[]
        self.bodies=[]
        for e in elems:
            e.set_aircraft(self)
            if type(e)==wing:
                self.wings+=[e]
            elif type(e)==body:
                self.bodies+=[e]
        #BE VERY AWARE OF THESE WARNINGS WHEN USINGG FUSELAGE-SEPARATED SEMI-WINGS AS WING OBJECTS
        if len(self.wings)!=0:
            S, mac, b=self.wings[0].calc_reference()
        else:
            S=1.0; mac=1.0; b=1.0
        if Sref==0.0:
            print('WARNING: no reference surface provided. Using first wing input as reference. Address to aircraft constructor\'s Sref argument for better definitions')
            self.Sref=S
        else:
            self.Sref=Sref
        if cref==0.0:
            print('WARNING: no reference chord provided. Using first wing input as reference. Address to aircraft constructor\'s cref argument for better definitions')
            self.cref=mac
        else:
            self.cref=cref
        if bref==0.0:
            print('WARNING: no reference span provided. Using first wing input as reference. Address to aircraft constructor\'s bref argument for better definitions')
            self.bref=b
        else:
            self.bref=bref
        self.AR=(self.bref**2)/self.Sref

        self.a=0.0
        self.b=0.0
        self.p=0.0
        self.q=0.0
        self.r=0.0
        self.Uinf=1.0
        self.M=0.0
        self.rho=1.225
        self.mu=1.72e-5
        self.g=9.81
        self.gamma=1.4

        self.CG=CG
        
        #defining available controls
        self.controlset={}
        for wng in self.wings:
            for wngqd in wng.wingquads:
                if wngqd.hascontrol():
                    for cname in wngqd.controls:
                        if not cname in self.controlset:
                            self.controlset[cname]=control_DOF()
                            wngqd.controls[cname].DOF=self.controlset[cname]

        self.sld=sld
        self.forcesavailable=False
        self.stabavailable=False
        self.massavailable=False
        self.haseqflatplate=False
        self.hascorrections=any([wng.hascorrection() for wng in self.wings])

        if echo:
            self.parameter_report()

        #defining plotting limits
        xkeypts=[]
        ykeypts=[]
        zkeypts=[]
        for wng in self.wings:
            for wngqd in wng.wingquads:
                xkeypts+=[wngqd.sect1.CA_position[0]-wngqd.sect1.c/4, wngqd.sect1.CA_position[0]+3*wngqd.sect1.c/4, \
                    wngqd.sect2.CA_position[0]-wngqd.sect2.c/4, wngqd.sect2.CA_position[0]+3*wngqd.sect2.c/4]
                ykeypts+=[wngqd.sect1.CA_position[1], wngqd.sect2.CA_position[1]]
                zkeypts+=[wngqd.sect1.CA_position[2], wngqd.sect2.CA_position[2]]
        for bdy in self.bodies:
            xkeypts+=[bdy.sections[i].center[0] for i in range(len(bdy.sections))]
            ykeypts+=[bdy.sections[i].center[1]-bdy.sections[i].Rs[i] for i in range(len(bdy.sections))]
            zkeypts+=[bdy.sections[i].center[2]-bdy.sections[i].Rs[i] for i in range(len(bdy.sections))]
            ykeypts+=[bdy.sections[i].center[1]+bdy.sections[i].Rs[i] for i in range(len(bdy.sections))]
            zkeypts+=[bdy.sections[i].center[2]+bdy.sections[i].Rs[i] for i in range(len(bdy.sections))]
        xmax=max(xkeypts)
        xmin=min(xkeypts)
        ymax=max(ykeypts)
        ymin=min(ykeypts)
        zmax=max(zkeypts)
        zmin=min(zkeypts)
        self.plotlim=max([abs(xmax), abs(ymax), abs(zmax), abs(xmin), abs(ymin), abs(zmin)])
    def hascontrol(self):
        return len(self.controlset)!=0
    def addwake(self, offset=1000.0, wakedisc=1, strategy=lambda x: ((np.exp(x)-1.0)/(exp(1)-1.0))**2):
        #once the wake has been added, one can end geometry pre-processing of panels in question:
        self.sld.end_preprocess()
        for wng in self.wings:
            wng.genwakepanels(offset=offset, a=self.a, b=self.b, wakedisc=wakedisc, strategy=strategy)
    def calcforces(self, echo=True):
        self.CX=sum([self.sld.forces[i][0] for i in range(self.sld.npanels)])/self.Sref
        self.CY=sum([self.sld.forces[i][1] for i in range(self.sld.npanels)])/self.Sref
        self.CZ=sum([self.sld.forces[i][2] for i in range(self.sld.npanels)])/self.Sref
        u=np.array([cos(self.a)*cos(self.b), sin(self.b)*cos(self.a), sin(self.a)])
        v=np.array([-sin(self.a)*cos(self.b), -sin(self.b)*sin(self.a), cos(self.a)])
        adforces=np.array([self.CX, self.CY, self.CZ])
        self.CL=v@adforces
        self.CD=u@adforces
        self.Cl=sum([self.sld.moments[i][0] for i in range(self.sld.npanels)])/(self.Sref*self.bref)
        self.Cm=sum([self.sld.moments[i][1] for i in range(self.sld.npanels)])/(self.Sref*self.cref)
        self.Cn=sum([self.sld.moments[i][2] for i in range(self.sld.npanels)])/(self.Sref*self.bref)

        if self.hascorrections:
            self.dCX=0.0; self.dCY=0.0; self.dCZ=0.0; self.dCL=0.0; self.dCD=0.0; self.dCl=0.0; self.dCm=0.0; self.dCn=0.0
            for wng in self.wings:
                wng.calc_coefs(alpha=self.a, axis=wng.axis)
                if wng.hascorrection():
                    dCX_wng, dCY_wng, dCZ_wng, dCl_wng, dCm_wng, dCn_wng=wng.calc_corrected_forces()
                    self.dCX+=dCX_wng
                    self.dCY+=dCY_wng
                    self.dCZ+=dCZ_wng
                    self.dCl+=dCl_wng
                    self.dCm+=dCm_wng
                    self.dCn+=dCn_wng
            adforces=np.array([self.dCX, self.dCY, self.dCZ])
            self.dCL=adforces@v
            self.dCD=adforces@u
        
        self.transp_byvec(self.CG)

        if echo:
            self.forces_report()
        self.forcesavailable=True
    def freestream_derivative(self, par='a'): #definition of derivative of freestream velocity vectors u and v (normal vectors)
        if par=='a':
            du=np.array([-sin(self.a)*cos(self.b), -sin(self.b)*sin(self.a), cos(self.a)])
            dv=np.array([-cos(self.a)*cos(self.b), -sin(self.b)*cos(self.a), -sin(self.a)])
        elif par=='b':
            du=np.array([-cos(self.a)*sin(self.b), cos(self.b)*cos(self.a), 0.0])
            dv=np.array([sin(self.a)*sin(self.b), -cos(self.b)*sin(self.a), 0.0])
        else:
            du=np.zeros(3)
            dv=np.zeros(3)
        return du, dv
    def calcstab(self, echo=True):
        #calculate variation on Cps in respect to several parameters
        pars=['a', 'b', 'p', 'q', 'r']
        if self.forcesavailable:
            self.transp_byvec(-self.CG)
        self.stabderivative_dict={}
        orforces=np.array([self.CX, self.CY, self.CZ])
        for p in pars:
            dvdksi=self.sld.gen_farfield_derivative(self.Uinf, a=self.a, b=self.b, p=self.p*2*self.Uinf/self.bref, q=self.q*2*self.Uinf/self.cref, r=self.r*2*self.Uinf/self.bref, par=p)
            dCps=self.sld.calc_derivative_dv(self.Uinf, dvdksi)
            forces=[-dCps[i]*self.sld.panels[i].S*self.sld.panels[i].nvector for i in range(self.sld.npanels)]
            moments=[np.cross(self.sld.panels[i].colpoint, forces[i]) for i in range(self.sld.npanels)]
            dCX=sum([forces[i][0] for i in range(self.sld.npanels)])/self.Sref
            dCY=sum([forces[i][1] for i in range(self.sld.npanels)])/self.Sref
            dCZ=sum([forces[i][2] for i in range(self.sld.npanels)])/self.Sref
            u=np.array([cos(self.a)*cos(self.b), sin(self.b)*cos(self.a), sin(self.a)])
            v=np.array([-sin(self.a)*cos(self.b), -sin(self.b)*sin(self.a), cos(self.a)])
            adforces=np.array([dCX, dCY, dCZ])
            du, dv=self.freestream_derivative(par=p)
            dCL=v@adforces+orforces@dv
            dCD=u@adforces+orforces@du
            dCl=sum([moments[i][0] for i in range(self.sld.npanels)])/(self.Sref*self.bref)
            dCm=sum([moments[i][1] for i in range(self.sld.npanels)])/(self.Sref*self.cref)
            dCn=sum([moments[i][2] for i in range(self.sld.npanels)])/(self.Sref*self.bref)
            self.stabderivative_dict[p]={'dCX':dCX, 'dCY':dCY, 'dCZ':dCZ, 'dCL':dCL, 'dCD':dCD, 'dCl':dCl, 'dCm':dCm, 'dCn':dCn}
        self.stabavailable=True
        self.transp_byvec(self.CG)
        if echo:
            self.stabreport()
    def stabreport(self): #report calculated stability derivatives, or calculate them in place
        if not self.stabavailable:
            self.calcstab(echo=False)
        print('Stability derivatives')
        print('%5s | %10s %10s %10s %10s %10s %10s %10s %10s' % ('par', 'dCX', 'dCY', 'dCZ', 'dCL', 'dCD', 'dCl', 'dCm', 'dCn'))
        for p in self.stabderivative_dict:
            print('%5s | %10f %10f %10f %10f %10f %10f %10f %10f' % (p, self.stabderivative_dict[p]['dCX'], self.stabderivative_dict[p]['dCY'], \
                self.stabderivative_dict[p]['dCZ'], self.stabderivative_dict[p]['dCL'], self.stabderivative_dict[p]['dCD'], -self.stabderivative_dict[p]['dCl'], \
                    self.stabderivative_dict[p]['dCm'], self.stabderivative_dict[p]['dCn']))
    def edit_parameters(self, pardict, echo=True):
        for par in pardict:
            val=pardict[par]
            if par=='a':
                self.a=radians(val)
            elif par=='b':
                self.b=radians(val)
            elif par=='p':
                self.p=val
            elif par=='q':
                self.q=val
            elif par=='r':
                self.r=val
            elif par=='Uinf':
                self.Uinf=val
            elif par=='M':
                self.M=val
            elif par=='rho':
                self.rho=val
            elif par=='mu':
                self.mu=val
            elif par=='g':
                self.g=val
            elif par=='gamma':
                self.gamma=val
            else:
                self.controlset[par].state=radians(val)
        #reset result readiness
        self.stabavailable=False
        self.forcesavailable=False
        self.haseqflatplate=False
        if echo:
            self.parameter_report()
    def bodies_eqflatplate_apply(self, rho=1.225, mu=1.72*10e-5, turbulent_criterion=Re2e5, Cf_l_rule=Blausius_Cf_l, Cf_t_rule=Prandtl_1_7th):
        self.haseqflatplate=True
        for bdy in self.bodies:
            bdy.apply_eqflatplate(rho=rho, mu=mu, turbulent_criterion=turbulent_criterion, Cf_l_rule=Cf_l_rule, Cf_t_rule=Cf_t_rule, Uinf=self.Uinf)
    def eulersolve(self, echo=True, damper=0.0, tolerance=1e-5, wakeiter=0):
        self.sld.eulersolve(target=np.array([]), a=self.a, b=self.b, p=self.p*2*self.Uinf/self.bref, q=self.q*2*self.Uinf/self.cref, \
            r=self.r*2*self.Uinf/self.bref, damper=damper, Uinf=self.Uinf, echo=True, M=self.M, gamma=self.gamma, tolerance=tolerance, wakeiter=wakeiter, beta=sqrt(1.0-self.M**2))
    def forces_report(self):
        print('========Total Forces Report========')
        if self.hascorrections:
            if self.haseqflatplate:
                if self.haseqflatplate:
                    str_Cf='+local Cf estimation'
            else:
                str_Cf=''
            print('Inviscid'+str_Cf+' case: ')
        if not self.forcesavailable:
            self.calcforces(echo=False)
        print('%10s %10s %10s %10s %10s %10s | %10s %10s' % ('CX', 'CY', 'CZ', 'Cl', 'Cm', 'Cn', 'CL', 'CD'))
        print('%10f %10f %10f %10f %10f %10f | %10f %10f' % (self.CX, self.CY, self.CZ, -self.Cl, self.Cm, -self.Cn, self.CL, self.CD))
        if self.hascorrections:
            print('After viscous external polar corrections: ')
            print('%10s %10s %10s %10s %10s %10s | %10s %10s' % ('CX', 'CY', 'CZ', 'Cl', 'Cm', 'Cn', 'CL', 'CD'))
            print('%10f %10f %10f %10f %10f %10f | %10f %10f' % (self.CX+self.dCX, self.CY+self.dCY, self.CZ+self.dCZ, \
                -(self.Cl+self.dCl), self.Cm+self.dCm, -(self.Cn+self.dCn), self.CL+self.dCL, self.CD+self.dCD))
        print('===================================')
    def plot_input(self, xlim=[], ylim=[], zlim=[]):
        fig=plt.figure()
        ax=plt.axes(projection='3d')
        for elem in self.bodies+self.wings:
            elem.plot_input(ax=ax, fig=fig)
        if len(xlim)!=0:
            ax.set_xlim3d(xlim[0], xlim[1])
        else:
            ax.set_xlim3d(-self.plotlim, self.plotlim)
        if len(ylim)!=0:
            ax.set_ylim3d(ylim[0], ylim[1])
        else:
            ax.set_ylim3d(-self.plotlim, self.plotlim)
        if len(zlim)!=0:
            ax.set_zlim3d(zlim[0], zlim[1])
        else:
            ax.set_zlim3d(-self.plotlim, self.plotlim)
        plt.xlabel('x')
        plt.ylabel('y')
        plt.show()
    def plotgeometry(self, xlim=[], ylim=[], zlim=[], factor=1.0, velfield=True):
        if len(xlim)==0:
            xlim=[-self.plotlim, self.plotlim]
        if len(ylim)==0:
            ylim=[-self.plotlim, self.plotlim]
        if len(zlim)==0:
            zlim=[-self.plotlim, self.plotlim]
        self.sld.plotgeometry(xlim=xlim, ylim=ylim, zlim=zlim, factor=factor, velfield=velfield)
    def design_derivatives(self, wings=[], echo=True): #calculate derivatives in forces due to design parameters in a wing - its sectional incidences
        if len(wings)==0:
            wings=self.wings
        wnglist=[]
        u=np.array([cos(self.a)*cos(self.b), sin(self.b)*cos(self.a), sin(self.a)])
        v=np.array([-sin(self.a)*cos(self.b), -sin(self.b)*sin(self.a), cos(self.a)])
        wngcnt=1
        for wng in wings:
            sectcnt=1
            wnglist+=[[]]
            if echo:
                print('=============================================')
                print('Design derivatives for incidences in wing %d' % (wngcnt))
            for sect in range(len(wng.wingquads)+1):
                dFs, dMs=wng.calc_design_dFM(section=sect, Uinf=self.Uinf, rho=self.rho, CG=self.CG)
                #transform to derivatives in respect to degrees
                dFs*=np.pi/180.0
                dMs*=np.pi/180.0
                dFs/=self.Sref*self.rho*self.Uinf**2/2
                dMs[0]/=self.Sref*self.bref*self.rho*self.Uinf**2/2
                dMs[1]/=self.Sref*self.cref*self.rho*self.Uinf**2/2
                dMs[2]/=self.Sref*self.bref*self.rho*self.Uinf**2/2
                stwise=np.array([dFs@u, dFs@v])
                derdict={'dCX':dFs[0], 'dCY':dFs[1], 'dCZ':dFs[2], 'dCD':stwise[0], 'dCL':stwise[1], 'dCl':-dMs[0], 'dCm':dMs[1], 'dCn':-dMs[2]} #stability derivatives
                #transformed to derivative coordinate system
                if echo:
                    print('===========Section %d==========' % (sectcnt))
                    print('%8s %8s %8s %8s %8s %8s %8s %8s' % ('dCX', 'dCY', 'dCZ', 'dCD', 'dCL', 'dCl', 'dCm', 'dCn'))
                    print('%8f %8f %8f %8f %8f %8f %8f %8f' % (derdict['dCX'], derdict['dCY'], derdict['dCZ'], derdict['dCD'], derdict['dCL'], \
                        derdict['dCl'], derdict['dCm'], derdict['dCn']))
                wnglist[-1]+=[derdict]
                sectcnt+=1
            if echo:
                print('=============================================')
            wngcnt+=1
        return wnglist
            


import numpy as np
import numpy.linalg as lg
import scipy.linalg as slg
import scipy.linalg.lapack as slapack
import scipy.sparse.linalg as splg
from math import *
import matplotlib.pyplot as plt
from mpl_toolkits import mplot3d
import scipy.sparse as sps
import time as tm
import multiprocessing as mp
import os

import toolkit
from utils import *

class Panel: #Panel data type
    def __init__(self, lines):
        self.lines=lines
        self.wakelines=[]
        self.no_selfinf=[]
    def nocirc_enforce(self, linind):
        if linind!=-1:
            self.no_selfinf+=[linind]

def subprocess_genaicm(queue1, queue2): #generic function to unpack AICM calc order from multiprocessing 
    #queue and deliver it to calculating function in toolkit
    order=queue1.get()
    AICM=toolkit.aicm_lines_gen(order[2], order[3])
    queue2.put((order[0], order[1], AICM))

class WakeLine: #class encompassing info about a wake line. Made so as to compute deformation with wake rollup
    def __init__(self, ind): #recieves signed index as input
        self.ind=ind
        self.v=np.zeros(3)
        self.nabut=0
        self.updateme=True
        self.inverted=False
    def addvel(self, v): #add velocity to be averaged
        self.v+=v
        self.nabut+=1

class WakePanel:
    def __init__(self, wl, wr, center): #add a wake panel based on adjacent lines
        self.wl=wl
        self.wr=wr
        self.center=center
        self.v=np.zeros(3)

class Solid:
    def __init__(self, sldlist=[], wraparounds=[]): #solid data type
        self.panels=[]
        self.lines=[]
        self.addto=[]
        self.wakestrips=[]
        self.wakeline_inds=set([])
        self.wakelines=set([])
        self.problematic=[]
        self.npanels=0
        self.nlines=0
        self.nwake=0 #number of wake panels
        self.solavailable=False
        if len(wraparounds)==0:
            wraparounds=[[]]*len(sldlist)
        sldcnt=0
        for sld in sldlist: #define first patches (without their interconnections)
            wraps=wraparounds[sldcnt]
            self.addpatch(sld, wraps=wraps)
            sldcnt+=1
    def addpatch(self, sld, wraps=[], prevlines={}, invlats=[], tolerance=0.00005): #add panel based on point grid (list of lists)
        # returns: arrays of, respectively, horizontal and vertical line indexes; panel indexes list of lists;
        # point grid inserted as input, for external reference from high-end functions
        # lines provided with -2 index in prevlines will be accounted as new lines, and as -1, 
        # as inexistent (dismissable length) lines
        horzlines=[]
        vertlines=[]
        paninds=[]
        for i in range(len(sld)-1):
            horzlines+=[[]]
            vertlines+=[[]]
            for j in range(len(sld[0])-1):
                if i==0 and 'low' in prevlines:
                    if prevlines['low'][j]!=-2:
                        horzlines[-1]+=[prevlines['low'][j]]
                    else:
                        horzlines[-1]+=[self.addline(np.vstack((sld[i][j], sld[i][j+1])).T, tolerance=tolerance)]
                else:
                    horzlines[-1]+=[self.addline(np.vstack((sld[i][j], sld[i][j+1])).T, tolerance=tolerance)]
                if j==0 and 'right' in prevlines:
                    if prevlines['right'][i]!=-2:
                        vertlines[-1]+=[prevlines['right'][i]]
                    else:
                        vertlines[-1]+=[self.addline(np.vstack((sld[i][j], sld[i+1][j])).T, tolerance=tolerance)]
                else:
                    vertlines[-1]+=[self.addline(np.vstack((sld[i][j], sld[i+1][j])).T, tolerance=tolerance)]
        if 0 in wraps: #if wrapped in x direction (inner list layer), repeat the first lateral vertical lines
            for i in range(len(sld)-1):
                vertlines[i]+=[vertlines[i][0]]
        else:
            if 'left' in prevlines:
                for i in range(len(sld)-1):
                    if prevlines['left'][i]!=-2:
                        vertlines[i]+=[prevlines['left'][i]]
                    else:
                        vertlines[i]+=[self.addline(np.vstack((sld[i][-1], sld[i+1][-1])).T, tolerance=tolerance)]
            else:
                for i in range(len(sld)-1):
                    vertlines[i]+=[self.addline(np.vstack((sld[i][-1], sld[i+1][-1])).T, tolerance=tolerance)]
        horzlines+=[[]]
        if 1 in wraps: #same for y axis
            for i in range(len(sld[0])-1):
                horzlines[-1]+=[horzlines[0][i]]
        else:
            if 'up' in prevlines:
                for i in range(len(sld[0])-1):
                    if prevlines['up'][i]!=-2:
                        horzlines[-1]+=[prevlines['up'][i]]
                    else:
                        horzlines[-1]+=[self.addline(np.vstack((sld[-1][i], sld[-1][i+1])).T, tolerance=tolerance)]
            else:
                for i in range(len(sld[0])-1):
                    horzlines[-1]+=[self.addline(np.vstack((sld[-1][i], sld[-1][i+1])).T, tolerance=tolerance)]
        for i in range(len(sld)-1):
            paninds+=[[]]
            for j in range(len(sld[0])-1):
                l=[]
                if j==len(sld[0])-2 and 'left' in invlats:
                    l+=[3]
                if j==0 and 'right' in invlats:
                    l+=[1]
                if i==0 and 'low' in invlats:
                    l+=[0]
                if i==len(sld)-2 and 'up' in invlats:
                    l+=[2]
                paninds[-1]+=[self.addpanel([horzlines[i][j], vertlines[i][j+1], horzlines[i+1][j], vertlines[i][j]], invs=l)]
        return horzlines, vertlines, paninds, sld
    def addline(self, coords, tolerance=0.00005):
        #neglect lines of neglectible size
        if lg.norm(coords[:, 1]-coords[:, 0])>tolerance:
            if np.size(self.lines, 0)==0:
                self.lines=np.array([list(coords)])
            else:
                self.lines=np.append(self.lines, np.array([list(coords)]), axis=0)
            self.nlines+=1
            return self.nlines-1
        else:
            return -1
    def addpanel(self, lines, invs=[]): #line list: includes indexes summed by one, inverted to indicate oposite direction
        #for vortex line segment
        self.npanels+=1
        llist=[]
        n=0
        for l in lines:
            if l!=-1:
                if (n==3 or n==2):
                    llist+=[-1-l]
                else:
                    llist+=[l+1]
                if n in invs:
                    llist[-1]=-llist[-1]
            n+=1
        self.panels+=[Panel(llist)]
        return self.npanels-1
    def addwakepanel(self, refup, refdown, indup=0, indown=2, offsetleft=np.array([1000.0, 0.0, 0.0]), offsetright=np.array([]), \
        tolerance=0.00005, disc=10, strategy=lambda x: ((np.exp(x)-1.0)/(exp(1)-1.0))**2, prevleft=[], leftinvert=False, prevright=[], rightinvert=False):
        #add line segments correspondent to panel wake, and remove "inup-th"/"indown-th" line segment in the panel's list
        if len(offsetright)==0:
            offsetright=offsetleft
        self.nwake+=1
        wakedisc=strategy(np.linspace(0.0, 1.0, disc+1))
        if indup<len(self.panels[refup].lines):
            te_linind=self.panels[refup].lines.pop(indup)
        else:
            te_linind=self.panels[refup].wakelines.pop(indup-len(self.panels[refup].lines))
        self.panels[refup].TE_line=te_linind
        coords=self.line_getcoords(te_linind)
        self.wakestrips+=[[]]
        indsup=[]
        leftlines=[]
        rightlines=[]
        if len(prevright)==0:
            for i in range(disc):
                p0=coords[:, 0]+offsetright*wakedisc[i]
                p1=coords[:, 0]+offsetright*wakedisc[i+1]
                indsup+=[self.addline(np.vstack((p0, p1)).T, tolerance=tolerance)+1]
                rightlines+=[WakeLine(indsup[-1])]
        else:
            if rightinvert:
                rightlines=[prevright[i] for i in range(len(prevright)-1, -1, -1)]
                indsup+=[-prevright[i].ind for i in range(len(prevright)-1, -1, -1)]
            else:
                indsup+=[prevright[i].ind for i in range(len(prevright))]
                rightlines=prevright
        if len(prevleft)==0:
            for i in range(disc, 0, -1):
                p0=coords[:, 1]+offsetleft*wakedisc[i]
                p1=coords[:, 1]+offsetleft*wakedisc[i-1]
                indsup+=[self.addline(np.vstack((p0, p1)).T, tolerance=tolerance)+1]
                leftlines+=[WakeLine(indsup[-1])]
        else:
            if leftinvert:
                leftlines=[prevleft[i] for i in range(len(prevleft)-1, -1, -1)]
                indsup+=[-prevleft[i].ind for i in range(len(prevleft)-1, -1, -1)]
            else:
                indsup+=[prevleft[i].ind for i in range(len(prevleft))]
                leftlines=prevleft
        self.panels[refup].wakelines+=indsup
        for l in leftlines+rightlines:
            self.wakelines.add(l)
        for lind in indsup:
            self.wakeline_inds.add(abs(lind)-1)
        if refdown!=-1:
            te_linind=self.panels[refdown].lines.pop(indown)
            self.panels[refdown].wakelines+=[-l for l in indsup]
            self.panels[refdown].TE_line=te_linind
        for i in range(disc):
            center=(self.line_midpoint(rightlines[i].ind)+self.line_midpoint(leftlines[disc-i-1].ind))/2
            self.wakestrips[-1]+=[WakePanel(leftlines[disc-i-1], rightlines[i], center)]
        return leftlines, rightlines #return left, right wake lines
    def genwakepanels(self, wakecombs=[], wakeinds=[], offset=1000.0, a=0.0, b=0.0, disc=10, strategy=lambda x: ((np.exp(x)-1.0)/(exp(1)-1.0))**2, \
        prevleft=[], prevright=[], leftinvert=False, rightinvert=False):
        #generate wake panels based in list of TE panel combinations
        #wakecombs: list of lists, first element in sublist is upper surface panel
        #wakeinds: list of lists indicating corresponding wake vortex line segment index to apply kutta condition to
        if len(wakeinds)==0:
            wakeinds=[[0, 2]]
        trimlist(len(wakecombs), wakeinds)
        self.addto+=[[wakecombs[0][0], wakecombs[0][1]]]
        left, prevleft=self.addwakepanel(wakecombs[0][0], wakecombs[0][1], indup=wakeinds[0][0], indown=wakeinds[0][0], \
            offsetleft=np.array([cos(a)*cos(b), -cos(a)*sin(b), sin(a)])*offset, disc=disc, strategy=strategy, prevleft=prevleft, leftinvert=leftinvert)
        for i in range(1, len(wakecombs)-1):
            self.addto+=[[wakecombs[i][0], wakecombs[i][1]]]
            _, prevleft=self.addwakepanel(wakecombs[i][0], wakecombs[i][1], indup=wakeinds[i][0], indown=wakeinds[i][0], \
                offsetleft=np.array([cos(a)*cos(b), -cos(a)*sin(b), sin(a)])*offset, disc=disc, strategy=strategy, \
                    prevleft=prevleft, leftinvert=True)
        self.addto+=[[wakecombs[-1][0], wakecombs[-1][1]]]
        _, prevleft=self.addwakepanel(wakecombs[-1][0], wakecombs[-1][1], indup=wakeinds[-1][0], indown=wakeinds[-1][0], \
            offsetleft=np.array([cos(a)*cos(b), -cos(a)*sin(b), sin(a)])*offset, disc=disc, strategy=strategy, \
                prevleft=prevleft, leftinvert=True, prevright=prevright, rightinvert=rightinvert)
        return left, prevleft
        #straight, alpha and beta defined single wake panel as default
    def panel_getcoords(self, p): #get coordinates for panel based on line vectors and lists
        if p.lines[0]<0:
            coords=self.lines[-1-p.lines[0], :, :].T
        else:
            coords=self.lines[p.lines[0]-1, :, :].T
        for i in range(1, len(p.lines)):
            if p.lines[i]<0:
                coords=np.vstack((coords, self.lines[-1-p.lines[i], :, 1]))
            else:
                coords=np.vstack((coords, self.lines[p.lines[i]-1, :, 1]))
        coords=coords.T
        if np.size(coords, 1)==3:
            temp=coords
            coords=np.zeros((3, 4))
            coords[:, 0:3]=temp
            coords[:, -1]=temp[:, -1]
        return coords
    def panel_calcSn(self, p): #update panel area and normal vector
        l=self.line_getcoords(p.lines[0])
        p.nvector=np.cross(l[:, 1]-p.colpoint, l[:, 0]-p.colpoint)
        p.S=lg.norm(p.nvector)
        p.nvector/=p.S
        p.S/=2
        for i in range(1, len(p.lines)):
            l=self.line_getcoords(p.lines[i])
            p.S+=lg.norm(np.cross(l[:, 1]-p.colpoint, l[:, 0]-p.colpoint))/2
    def line_getcoords(self, ind): #return line coordinates in order presented in panel, based on panel.lines element
        if ind>0:
            return self.lines[ind-1, :, :]
        else:
            return np.flip(self.lines[-1-ind, :, :], axis=1)
    def line_getvec(self, ind): #return vector linking points in line
        if ind>0:
            return self.lines[ind-1, :, 1]-self.lines[ind-1, :, 0]
        else:
            return self.lines[-ind-1, :, 0]-self.lines[-ind-1, :, 1]
    def line_midpoint(self, ind): #return midpoint of line
        return np.mean(self.lines[abs(ind)-1, :, :], axis=1)
    def nvect_diradjust(self, patchinds, vect): #adjust normal vectors in patch to follow a certain direction
        for patchlist in patchinds:
            for i in patchlist:
                if self.panels[i].nvector@vect<0.0:
                    self.panels[i].nvector*=-1
    def nvect_radadjust(self, patchinds, center, inwards=False): #adjust normal vectors in patch to follow radial convergent or 
        #divergent distribution
        if inwards:
            for patchlist in patchinds:
                for i in patchlist:
                    if self.panels[i].nvector@(self.panels[i].colpoint-center)>0.0:
                        self.panels[i].nvector*=-1
        else:
            for patchlist in patchinds:
                for i in patchlist:
                    if self.panels[i].nvector@(self.panels[i].colpoint-center)<0.0:
                        self.panels[i].nvector*=-1
    def lineadjust(self, patchinds=[]): #adjust line fortran index signals to comply with dextrogyre panel convention
        if len(patchinds)==0:
            patchinds=[list(range(self.npanels))]
        p=np.array([0.0, 0.0, 0.0])
        u=np.array([0.0, 0.0, 0.0])
        for patchlist in patchinds:
            for i in patchlist:
                for lind in range(len(self.panels[i].lines)):
                    p=self.lines[abs(self.panels[i].lines[lind])-1, :, 0]-self.panels[i].colpoint
                    u=self.lines[abs(self.panels[i].lines[lind])-1, :, 1]-self.lines[abs(self.panels[i].lines[lind])-1, :, 0]
                    if (u@np.cross(self.panels[i].nvector, p)*self.panels[i].lines[lind]>0.0):
                        self.panels[i].lines[lind]*=-1
                        print('WARNING: line '+str(abs(self.panels[i].lines[lind])-1)+' in panel '+\
                            str(i)+' had to be inverted. Please check integrity of patchcompose() functions')
                        self.problematic+=[abs(l)-1 for l in self.panels[i].lines]
    def iscontiguous(self, patchinds=[], tolerance=5e-5): #check if all panels in solid are closed quadrilaterals
        if len(patchinds)==0:
            patchinds=[list(range(self.npanels))]
        for patchlist in patchinds:
            for i in patchlist:
                u=self.line_getcoords(self.panels[i].lines[0])
                for lind in range(1, len(self.panels[i].lines)):
                    v=u
                    u=self.line_getcoords(self.panels[i].lines[lind])
                    if np.amax(np.abs(v[:, 1]-u[:, 0]))>tolerance:
                        print('WARNING: '+str(i)+' panel is not contiguous')
                        self.problematic+=[abs(l)-1 for l in self.panels[i].lines]
                        break
    def end_preprocess(self, paninds=[], tolerance=5e-5, initialize=True, check=True):
        #calculate panel areas before they are altered by wake generation. Must be run before it, and terrible
        #consequences may arise if done otherwise
        u=np.array([0.0, 0.0, 0.0])
        v=np.array([0.0, 0.0, 0.0])
        #analyse whether provided panel index list corresponds to a certain set or, as default, indicates all panels in self
        allpans=(len(paninds)==0) or (len(paninds)==self.npanels)
        if allpans:
            panlist=self.panels
        else:
            panlist=[self.panels[i] for i in paninds]
        n=0
        for p in panlist:
            '''coords=self.panel_getcoords(p)
            u=coords[:, 1]-coords[:, 0]
            v=coords[:, 2]-coords[:, 1]'''
            u=self.line_getvec(p.lines[0])
            v=self.line_getvec(p.lines[1])
            p.nvector=np.cross(v, u)
            p.S=0.0
            if len(p.lines)==3:
                p.S=lg.norm(p.nvector)/2
                p.nvector/=lg.norm(p.nvector)
                p.colpoint=(self.line_midpoint(p.lines[0])+self.line_midpoint(p.lines[1])+self.line_midpoint(p.lines[2]))/3
                #p.colpoint=(coords[:, 0]+coords[:, 1]+coords[:, 2])/3
            elif len(p.lines)==4:
                p.S=(lg.norm(p.nvector)+lg.norm(np.cross(self.line_getvec(p.lines[2]), self.line_getvec(p.lines[3]))))/2
                #p.S=(lg.norm(p.nvector)+lg.norm(np.cross(coords[:, 1]-coords[:, 2], coords[:, 3]-coords[:, 2])))/2
                p.nvector/=lg.norm(p.nvector)
                #p.colpoint=(coords[:, 0]+coords[:, 1]+coords[:, 2]+coords[:, 3])/4
                p.colpoint=(self.line_midpoint(p.lines[0])+self.line_midpoint(p.lines[1])+self.line_midpoint(p.lines[2])+\
                    self.line_midpoint(p.lines[3]))/4
            else:
                p.nvector=np.array([0.0, 0.0, 0.0])
                #p.colpoint=(coords[:, 0]+coords[:, 1])/2
                p.colpoint=(self.line_midpoint(p.lines[0])+self.line_midpoint(p.lines[1]))/2
                p.S=0.0
            if p.S<=tolerance**2:
                print('WARNING: panel '+str(n)+' displayed incoherent area '+str(p.S)+' with '+str(len(p.lines))+' lines')
            n+=1
        #initialize result vectors
        if initialize:
            self.delphi=np.array([[0.0, 0.0, 0.0]]*len(self.panels), dtype='double')
            self.vbar=np.array([[0.0, 0.0, 0.0]]*len(self.panels), dtype='double')
            self.nvv=np.array([0.0]*len(self.panels), dtype='double')
            self.solution=np.array([0.0]*len(self.panels), dtype='double')
            self.solavailable=False
            self.Cps=np.array([0.0]*len(self.panels), dtype='double')
            self.Cfs=np.zeros(len(self.panels), dtype='double')
            self.forces=[]
            self.moments=[]
        #adjust lines in case any is set inconsistently with respect to anti-clockwise convention in panel
        if check:
            self.lineadjust()
            #self.iscontiguous(tolerance=tolerance)
    def gen_panline(self): #generate panel-line correspondence matrix
        self.panline_matrix=np.zeros((self.nlines, self.npanels), dtype='double')
        for i in range(self.npanels):
            lininds=np.array(self.panels[i].lines+self.panels[i].wakelines)
            lintemp=lininds[lininds>0]
            self.panline_matrix[lintemp-1, i]=1.0
            lintemp=lininds[lininds<0]
            self.panline_matrix[-lintemp-1, i]=-1.0
        self.panline_matrix=sps.csr_matrix(self.panline_matrix)
    def addorder(self, queue, colmat, ind1, ind2): #add order to queue for subprocess to compute AICM from rows ind1 to ind2
        queue.put((ind1, ind2, self.lines[ind1:ind2, :, :], colmat))
    def genaicm(self): #call FORTRAN backend to generate AIC matrix
        colmat=np.zeros((self.npanels, 3))
        nvectmat=np.zeros((self.npanels, 3))
        for i in range(len(self.panels)):
            colmat[i, :]=self.panels[i].colpoint
            nvectmat[i, :]=self.panels[i].nvector
        ncpus=mp.cpu_count()
        calclims=np.linspace(0, self.nlines, ncpus+1)
        calclims=[int(c) for c in calclims]
        calclims=[[calclims[i], calclims[i+1]] for i in range(ncpus)]
        queue1=mp.Queue()
        queue2=mp.Queue()
        for i in range(len(calclims)):
            self.addorder(queue1, colmat, ind1=calclims[i][0], ind2=calclims[i][1])
        processes=[]
        for i in range(ncpus):
            processes+=[mp.Process(target=subprocess_genaicm, args=(queue1, queue2))]
        for p in processes:
            #p.daemon=True
            p.start()
        ordresults=[]
        for i in range(ncpus):
            ordresults+=[queue2.get()]
        for p in processes:
            p.join()
        self.aicm3_line=np.zeros((3, self.npanels, self.nlines))
        for ordresult in ordresults:
            self.aicm3_line[:, :, ordresult[0]:ordresult[1]]=ordresult[2]
        self.gen_panline()
        self.aicm3=np.zeros((3, self.npanels, self.npanels))
        for i in range(3):
            self.aicm3[i, :, :]=self.aicm3_line[i, :, :]@self.panline_matrix
        self.aicm=toolkit.aicm_norm_conv(self.aicm3, nvectmat)
    def gen_farfield(self, Uinf, a=0.0, b=0.0, p=0.0, q=0.0, r=0.0): #generate generic local freestream velocity dependant on parameters
        newvec=np.zeros((self.npanels, 3))
        for i in range(self.npanels):
            newvec[i, :]=Uinf*np.array([cos(a)*cos(b), -cos(a)*sin(b), sin(a)], dtype='double')+np.cross(np.array([p, -q, r], dtype='double'), self.panels[i].colpoint)
        return newvec
    def gen_wake_farfield(self, Uinf, a=0.0, b=0.0, p=0.0, q=0.0, r=0.0): #generate generic local freestream at wake panels
        for strip in self.wakestrips:
            for pan in strip:
                pan.v=Uinf*np.array([cos(a)*cos(b), -cos(a)*sin(b), sin(a)], dtype='double')+np.cross(np.array([p, -q, r], dtype='double'), pan.center)
    def gen_farfield_derivative(self, Uinf, a=0.0, b=0.0, p=0.0, q=0.0, r=0.0, par='a'): #generate generic local freestream velocity dependant on parameters
        newvec=np.zeros((self.npanels, 3))
        if par=='a':
            for i in range(self.npanels):
                newvec[i, :]=Uinf*np.array([-sin(a)*cos(b), sin(a)*sin(b), cos(a)], dtype='double')
        elif par=='b':
            for i in range(self.npanels):
                newvec[i, :]=Uinf*np.array([-cos(a)*sin(b), -cos(a)*cos(b), 0.0], dtype='double')
        elif par=='p':
            for i in range(self.npanels):
                newvec[i, :]=np.cross(np.array([1.0, 0.0, 0.0], dtype='double'), self.panels[i].colpoint)
        elif par=='q':
            for i in range(self.npanels):
                newvec[i, :]=np.cross(np.array([0.0, -1.0, 0.0], dtype='double'), self.panels[i].colpoint)
        elif par=='r':
            for i in range(self.npanels):
                newvec[i, :]=np.cross(np.array([0.0, 0.0, 1.0], dtype='double'), self.panels[i].colpoint)
        elif par=='Uinf':
            for i in range(self.npanels):
                newvec[i, :]=np.array([cos(a)*cos(b), -cos(a)*sin(b), sin(a)], dtype='double')
        return newvec
    def genvbar(self, Uinf, a=0.0, b=0.0, p=0.0, q=0.0, r=0.0): #generate freestream velocity vector. Angular velocities are raw (rad/s), not normalized by Uinf or dimensions
        self.vbar=self.gen_farfield(Uinf, a=a, b=b, p=p, q=q, r=r)
    def gennvv(self, beta=1.0): #Generate vector containing normal velocities
        for i in range(self.npanels):
            self.nvv[i]=self.panels[i].nvector@self.vbar[i, :]
    def gen_selfinf_mat(self): #generate self-influence matrix according to Srivastava's equations, with each line's vorticity as an input
        #so as to ease stability derivative calculation
        self.selfinf_mat_x=np.zeros((self.npanels, self.nlines), order='F')
        self.selfinf_mat_y=np.zeros((self.npanels, self.nlines), order='F')
        self.selfinf_mat_z=np.zeros((self.npanels, self.nlines), order='F')
        for i in range(self.npanels):
            linelist=[]
            nocirc_linelist=[]
            for l in self.panels[i].lines:
                if not int(abs(l))-1 in self.panels[i].no_selfinf:
                    linelist+=[int(abs(l))-1]
                else:
                    nocirc_linelist+=[int(abs(l))-1]
            for l in linelist:
                vdv=np.cross(self.lines[l, :, 1]-self.lines[l, :, 0], self.panels[i].nvector)/(4*self.panels[i].S)
                self.selfinf_mat_x[i, l]=vdv[0]
                self.selfinf_mat_y[i, l]=vdv[1]
                self.selfinf_mat_z[i, l]=vdv[2]
            for l in nocirc_linelist:
                vdv=np.cross(self.lines[l, :, 1]-self.lines[l, :, 0], self.panels[i].nvector)/(4*self.panels[i].S)
                vdv/=2
                self.selfinf_mat_x[i, l]=vdv[0]
                self.selfinf_mat_y[i, l]=vdv[1]
                self.selfinf_mat_z[i, l]=vdv[2]
        self.selfinf_mat_x=sps.csr_matrix(self.selfinf_mat_x)
        self.selfinf_mat_y=sps.csr_matrix(self.selfinf_mat_y)
        self.selfinf_mat_z=sps.csr_matrix(self.selfinf_mat_z)
    def gen_selfinf(self): #generate self-influence velocity according to Srivastava's equations
        self.delphi[:, 0]+=self.selfinf_mat_x@self.solution_line
        self.delphi[:, 1]+=self.selfinf_mat_y@self.solution_line
        self.delphi[:, 2]+=self.selfinf_mat_z@self.solution_line
    def solve(self, damper=0.0, target=np.array([]), wakeiter=0, Uinf=1.0, a=0.0, b=0.0, p=0.0, q=0.0, r=0.0, tolerance=1e-5, echo=True):
        #generate Euler solution. Inverts AIC matrix with Tikhonov regularization if "damper" is set to non-zero value.
        #self.iaicm=toolkit.aicm_inversion(self.aicm, damper)
        if len(target)==0:
            target=np.zeros(self.npanels) # target: nvv aimed at, for stability derivative computation and/or viscous-inviscid coupling

        if damper!=0.0:
            self.iaicm=slg.inv(self.aicm.T@self.aicm+damper*np.eye(self.npanels, self.npanels))@self.aicm.T
        else:
            self.iaicm=slg.inv(self.aicm)
        tango=target-self.nvv

        self.solution=self.iaicm@tango
        self.solution_line=self.panline_matrix@self.solution
        self.solavailable=True
        if wakeiter!=0:
            wakelininds=np.array(list(self.wakeline_inds))
            colmat=np.zeros((self.npanels, 3))
            nvectmat=np.zeros((self.npanels, 3))
            for i in range(len(self.panels)):
                colmat[i, :]=self.panels[i].colpoint
                nvectmat[i, :]=self.panels[i].nvector
            if echo:
                print('%18s | %5s %5s %5s' % ('Iteration', 'vcalc', 'roll', 'solve'))
        for i in range(wakeiter): #iterate for wake generation
            t0=tm.time()
            self.gen_wake_farfield(Uinf=Uinf, a=a, b=b, p=p, q=q, r=r)
            self.add_wakevels(tolerance=tolerance)
            tvel=tm.time()
            self.wake_rollup()
            trol=tm.time()
            self.aicm3_line[:, :, wakelininds]=toolkit.aicm_lines_recalc(wakelininds+1, self.lines, colmat)
            for j in range(3):
                self.aicm3[j, :, :]=self.aicm3_line[j, :, :]@self.panline_matrix
            self.aicm=toolkit.aicm_norm_conv(self.aicm3, nvectmat)
            if damper!=0.0:
                self.iaicm=slg.inv(self.aicm.T@self.aicm+damper*np.eye(self.npanels, self.npanels))@self.aicm.T
            else:
                self.iaicm=slg.inv(self.aicm)
            self.solution=self.iaicm@tango
            self.solution_line=self.panline_matrix@self.solution
            tsol=tm.time()
            print('%14s %3d | %5f %5f %5f' % ('Wake teration', i, tvel-t0, trol-tvel, tsol-trol))
            
        for i in range(3):
            self.delphi[:, i]=self.aicm3[i, :, :]@self.solution #compute local velocity due to disturbance field
        self.gen_selfinf()
    def calcpress(self, Uinf=1.0, gamma=1.4, M=0.0):
        if M==0.0:
            for i in range(self.npanels):
                self.Cps[i]=(Uinf**2-(self.delphi[i, :]+self.vbar[i, :])@(self.delphi[i, :]+self.vbar[i, :]))/Uinf**2
        else:
            for i in range(self.npanels):
                incomp=(Uinf**2-(self.delphi[i, :]+self.vbar[i, :])@(self.delphi[i, :]+self.vbar[i, :]))/Uinf**2 #incompressible value
                self.Cps[i]=2*((1.0+(gamma-1)*M**2*incomp/2)**(gamma/(gamma-1))-1)/(M**2*gamma)
    def calcforces(self): #compute force correspondent to unitary dynamic pressure on each panel
        self.forces=[-self.panels[i].S*self.panels[i].nvector*self.Cps[i]+self.panels[i].S*self.Cfs[i]*\
            (self.vbar[i, :]+self.delphi[i, :])/lg.norm(self.vbar[i, :]+self.delphi[i, :]) for i in range(self.npanels)]
        self.moments=[np.cross(self.panels[i].colpoint, self.forces[i]) for i in range(self.npanels)]
    def calc_derivative_dv(self, Uinf, dvdksi, M=0.0, gamma=1.4): #calculate local Cp derivative by freestream derivative
        dndksi=np.array([self.panels[i].nvector@dvdksi[i, :] for i in range(self.npanels)])
        dGammadksi=-self.iaicm@dndksi
        dGamma_linedksi=self.panline_matrix@dGammadksi
        dvdksi[:, 0]+=self.aicm3[0, :, :]@dGammadksi+self.selfinf_mat_x@dGamma_linedksi
        dvdksi[:, 1]+=self.aicm3[1, :, :]@dGammadksi+self.selfinf_mat_y@dGamma_linedksi
        dvdksi[:, 2]+=self.aicm3[2, :, :]@dGammadksi+self.selfinf_mat_z@dGamma_linedksi
        dCps=np.array([-(2*(self.vbar[i, :]+self.delphi[i, :])@dvdksi[i, :])/Uinf**2 for i in range(self.npanels)])
        if M!=0.0:
            for i in range(self.npanels):
                incomp=(Uinf**2-(self.delphi[i, :]+self.vbar[i, :])@(self.delphi[i, :]+self.vbar[i, :]))/Uinf**2 #incompressible value
                dCps[i]=2*((1+(gamma-1)*M**2*incomp/2)**(1.0/(gamma-1.0)))*dCps[i]/(M**2*(gamma-1.0))
        return dCps
    def calc_derivative_dn(self, Uinf, dndksi, M=0.0, gamma=1.4): #calculate local Cp derivative by normal velocity derivative
        dGammadksi=-self.iaicm@dndksi
        dGamma_linedksi=self.panline_matrix@dGammadksi
        dvdksi=np.zeros((len(dndksi), 3))
        dvdksi[:, 0]=self.aicm3[0, :, :]@dGammadksi+self.selfinf_mat_x@dGamma_linedksi
        dvdksi[:, 1]=self.aicm3[1, :, :]@dGammadksi+self.selfinf_mat_y@dGamma_linedksi
        dvdksi[:, 2]=self.aicm3[2, :, :]@dGammadksi+self.selfinf_mat_z@dGamma_linedksi
        dCps=np.array([-(2*(self.vbar[i, :]+self.delphi[i, :])@dvdksi[i, :])/Uinf**2 for i in range(self.npanels)])
        if M!=0.0:
            for i in range(self.npanels):
                incomp=(Uinf**2-(self.delphi[i, :]+self.vbar[i, :])@(self.delphi[i, :]+self.vbar[i, :]))/Uinf**2 #incompressible value
                dCps[i]=2*((1+(gamma-1)*M**2*incomp/2)**(1.0/(gamma-1.0)))*dCps[i]/(M**2*(gamma-1.0))
        return dCps
    def add_wakevels(self, tolerance=1e-5): #calculate velocities at wake panel control points and add to local velocity variables
        for strip in self.wakestrips:
            for p in strip:
                p.v+=toolkit.get_field_influence(self.lines, self.solution_line, p.center, tolerance=tolerance)
    def wake_rollup(self): #calculate velocity at nodes located at wake line midpoints and update the lines
        for strip in self.wakestrips:
            for p in strip:
                p.wl.addvel(p.v)
                p.wr.addvel(p.v)
        self.wakelines_deform()
    def wakelines_deform(self): #deform wake lines according to computed velocity at panels between them
        for l in self.wakelines:
            l.v/=l.nabut
            l.inverted=self.lines[abs(l.ind)-1, 0, 0]>self.lines[abs(l.ind)-1, 0, 1]
            l.updateme=True
        deltaxl=0.0
        deltaxr=0.0
        dvl=np.zeros(3)
        dvr=np.zeros(3)
        origin_left=np.zeros(3)
        origin_right=np.zeros(3)
        newleft=np.zeros(3)
        newright=np.zeros(3)
        for strip in self.wakestrips:
            origin_left=self.lines[abs(strip[0].wl.ind)-1, :, 1 if strip[0].wl.inverted else 0]
            origin_right=self.lines[abs(strip[0].wr.ind)-1, :, 1 if strip[0].wr.inverted else 0]
            for i in range(len(strip)):
                if strip[i].wl.updateme:
                    deltaxl=abs(self.lines[abs(strip[i].wl.ind)-1, 0, 1]-self.lines[abs(strip[i].wl.ind)-1, 0, 0])
                    dvl=strip[i].wl.v*deltaxl/(strip[i].wl.v[0])
                    newleft=origin_left+dvl
                    if strip[i].wl.inverted:
                        self.lines[abs(strip[i].wl.ind)-1, :, 1]=origin_left
                        self.lines[abs(strip[i].wl.ind)-1, :, 0]=newleft
                    else:
                        self.lines[abs(strip[i].wl.ind)-1, :, 0]=origin_left
                        self.lines[abs(strip[i].wl.ind)-1, :, 1]=newleft
                    origin_left=newleft
                    strip[i].wl.updateme=False
                if strip[i].wr.updateme:
                    deltaxr=abs(self.lines[abs(strip[i].wr.ind)-1, 0, 1]-self.lines[abs(strip[i].wr.ind)-1, 0, 0])
                    dvr=strip[i].wr.v*deltaxr/(strip[i].wr.v[0])
                    newright=origin_right+dvr
                    if strip[i].wr.inverted:
                        self.lines[abs(strip[i].wr.ind)-1, :, 1]=origin_right
                        self.lines[abs(strip[i].wr.ind)-1, :, 0]=newright
                    else:
                        self.lines[abs(strip[i].wr.ind)-1, :, 0]=origin_right
                        self.lines[abs(strip[i].wr.ind)-1, :, 1]=newright
                    origin_right=newright
                    strip[i].wr.updateme=False
                strip[i].center=(self.line_midpoint(strip[i].wr.ind)+self.line_midpoint(strip[i].wl.ind))/2
        for l in self.wakelines:
            l.v=0.0
            l.nabut=0
    def PG_apply(self, beta, a, b): #beta indicates PG correction factor for compressibility
        PG_apmat=PG_xmult(beta, a, b)
        #convert lines
        self.lines[:, 0, :]=PG_apmat[0, 0]*self.lines[:, 0, :]+PG_apmat[0, 1]*self.lines[:, 1, :]+PG_apmat[0, 2]*self.lines[:, 2, :]
        self.lines[:, 1, :]=PG_apmat[1, 0]*self.lines[:, 0, :]+PG_apmat[1, 1]*self.lines[:, 1, :]+PG_apmat[1, 2]*self.lines[:, 2, :]
        self.lines[:, 2, :]=PG_apmat[2, 0]*self.lines[:, 0, :]+PG_apmat[2, 1]*self.lines[:, 1, :]+PG_apmat[2, 2]*self.lines[:, 2, :]
        #convert panels
        for p in self.panels:
            p.colpoint=PG_apmat@p.colpoint
            self.panel_calcSn(p)
        #convert wake panels
        for strip in self.wakestrips:
            for p in strip:
                p.center=PG_apmat@p.center
    def PG_remove(self, beta, a, b): #remove effect of previously imposed PG correction
        PG_apmat=PG_inv_xmult(beta, a, b)
        PG_apv=PG_vtouni(beta, a, b)
        #convert lines
        self.lines[:, 0, :]=PG_apmat[0, 0]*self.lines[:, 0, :]+PG_apmat[0, 1]*self.lines[:, 1, :]+PG_apmat[0, 2]*self.lines[:, 2, :]
        self.lines[:, 1, :]=PG_apmat[1, 0]*self.lines[:, 0, :]+PG_apmat[1, 1]*self.lines[:, 1, :]+PG_apmat[1, 2]*self.lines[:, 2, :]
        self.lines[:, 2, :]=PG_apmat[2, 0]*self.lines[:, 0, :]+PG_apmat[2, 1]*self.lines[:, 1, :]+PG_apmat[2, 2]*self.lines[:, 2, :]
        #convert panels
        for p in self.panels:
            p.colpoint=PG_apmat@p.colpoint
            self.panel_calcSn(p)
        #convert wake panels
        for strip in self.wakestrips:
            for p in strip:
                p.center=PG_apmat@p.center
        #convert self influence and aicm3_line matrixes
        self.selfinf_mat_x=PG_apv[0, 0]*self.selfinf_mat_x+PG_apv[0, 1]*self.selfinf_mat_y+PG_apv[0, 2]*self.selfinf_mat_z
        self.selfinf_mat_y=PG_apv[1, 0]*self.selfinf_mat_x+PG_apv[1, 1]*self.selfinf_mat_y+PG_apv[1, 2]*self.selfinf_mat_z
        self.selfinf_mat_z=PG_apv[2, 0]*self.selfinf_mat_x+PG_apv[2, 1]*self.selfinf_mat_y+PG_apv[2, 2]*self.selfinf_mat_z
        #self.aicm remains the same
        self.aicm3_line[0, :, :]=PG_apv[0, 0]*self.aicm3_line[0, :, :]+PG_apv[0, 1]*self.aicm3_line[1, :, :]+PG_apv[0, 2]*self.aicm3_line[2, :, :]
        self.aicm3_line[1, :, :]=PG_apv[1, 0]*self.aicm3_line[0, :, :]+PG_apv[1, 1]*self.aicm3_line[1, :, :]+PG_apv[1, 2]*self.aicm3_line[2, :, :]
        self.aicm3_line[2, :, :]=PG_apv[2, 0]*self.aicm3_line[0, :, :]+PG_apv[2, 1]*self.aicm3_line[1, :, :]+PG_apv[2, 2]*self.aicm3_line[2, :, :]
        self.aicm3[0, :, :]=PG_apv[0, 0]*self.aicm3[0, :, :]+PG_apv[0, 1]*self.aicm3[1, :, :]+PG_apv[0, 2]*self.aicm3[2, :, :]
        self.aicm3[1, :, :]=PG_apv[1, 0]*self.aicm3[0, :, :]+PG_apv[1, 1]*self.aicm3[1, :, :]+PG_apv[1, 2]*self.aicm3[2, :, :]
        self.aicm3[2, :, :]=PG_apv[2, 0]*self.aicm3[0, :, :]+PG_apv[2, 1]*self.aicm3[1, :, :]+PG_apv[2, 2]*self.aicm3[2, :, :]
    def plotgeometry(self, xlim=[], ylim=[], zlim=[], velfield=True, factor=1.0):
        #plot geometry and local velocity vectors, either with or without wake panels
        fig=plt.figure()
        ax=plt.axes(projection='3d')
        for i in range(self.nlines):
            ax.plot3D(self.lines[i, 0, :], self.lines[i, 1, :], self.lines[i, 2, :], 'gray')
        for i in self.problematic:
            ax.plot3D(self.lines[i, 0, :], self.lines[i, 1, :], self.lines[i, 2, :], 'red')
        if self.solavailable and velfield:
            '''ax.quiver([p.colpoint[0] for p in self.panels], [p.colpoint[1] for p in self.panels], \
                [p.colpoint[2] for p in self.panels], [p.nvector[0]*0.005 for p in self.panels], \
                    [p.nvector[1]*0.005 for p in self.panels], \
                        [p.nvector[2]*0.005 for p in self.panels])'''
            ax.quiver([p.colpoint[0] for p in self.panels], [p.colpoint[1] for p in self.panels], \
                [p.colpoint[2] for p in self.panels], [(self.vbar[i, 0]+self.delphi[i, 0])*factor for i in range(self.npanels)], \
                    [(self.vbar[i, 1]+self.delphi[i, 1])*factor for i in range(self.npanels)], \
                        [(self.vbar[i, 2]+self.delphi[i, 2])*factor for i in range(self.npanels)])
        if len(xlim)!=0:
            ax.set_xlim3d(xlim[0], xlim[1])
        if len(ylim)!=0:
            ax.set_ylim3d(ylim[0], ylim[1])
        if len(zlim)!=0:
            ax.set_zlim3d(zlim[0], zlim[1])
        ax.view_init(azim=-135, elev=30)
        plt.xlabel('x')
        plt.ylabel('y')
        plt.show()
    def plotnormals(self, xlim=[], ylim=[], zlim=[], factor=1.0):
        #plot normal vectors of panels. Function essentially produced for geometry gen. debugging
        fig=plt.figure()
        ax=plt.axes(projection='3d')
        for i in range(self.nlines):
            ax.plot3D(self.lines[i, 0, :], self.lines[i, 1, :], self.lines[i, 2, :], 'gray')
        ax.quiver([p.colpoint[0] for p in self.panels], [p.colpoint[1] for p in self.panels], \
            [p.colpoint[2] for p in self.panels], [p.nvector[0]*factor for p in self.panels], \
                [p.nvector[1]*factor for p in self.panels], \
                    [p.nvector[2]*factor for p in self.panels])
        '''ax.quiver([p.colpoint[0] for p in self.panels], [p.colpoint[1] for p in self.panels], \
            [p.colpoint[2] for p in self.panels], [self.vbar[i, 0]+self.delphi[i, 0] for i in range(self.npanels)], \
                [self.vbar[i, 1]+self.delphi[i, 1] for i in range(self.npanels)], \
                    [self.vbar[i, 2]+self.delphi[i, 2] for i in range(self.npanels)])'''
        if len(xlim)!=0:
            ax.set_xlim3d(xlim[0], xlim[1])
        if len(ylim)!=0:
            ax.set_ylim3d(ylim[0], ylim[1])
        if len(zlim)!=0:
            ax.set_zlim3d(zlim[0], zlim[1])
        ax.view_init(azim=-135, elev=30)
        plt.xlabel('x')
        plt.ylabel('y')
        plt.show()
    def eulersolve(self, target=np.array([]), Uinf=1.0, M=0.0, gamma=1.4, beta=1.0, a=0.0, b=0.0, p=0.0, q=0.0, r=0.0, damper=0.0, echo=True, \
        wakeiter=0, tolerance=1e-5):
        if echo:
            print('========Euler solution=======')
            print(self.npanels, ' panels, ', self.nlines, ' lines')
        t=tm.time()
        self.genvbar(Uinf, a=a, b=b, p=p, q=q, r=r)
        self.gennvv()
        self.PG_apply(beta, a, b)
        if echo:
            print('Pre-processing: '+str(tm.time()-t)+' s')
        t=tm.time()
        self.genaicm()
        self.gen_selfinf_mat()
        if echo:
            print('AIC matrix generation: '+str(tm.time()-t)+' s')
        t=tm.time()
        self.PG_remove(beta, a, b)
        self.solve(damper=damper, wakeiter=wakeiter, Uinf=Uinf, a=a, b=b, p=p, q=q, r=r, tolerance=tolerance, echo=echo)
        self.calcpress(Uinf=Uinf, M=M, gamma=gamma)
        self.calcforces()
        if echo:
            print('Solution and post-processing: '+str(tm.time()-t)+' s')
            print('=============================')

'''sld=Solid(sldlist=[np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [1.0, 1.0, 0.0], [0.0, 1.0, 0.0]]).T])
#sld.plotgeometry()
sld.genaicm()
print(sld.aicm3[:, 0, 0])'''


import numpy as np
import numpy.linalg as lg
import scipy.linalg as slg
import scipy.linalg.lapack as slapack
from math import *
import matplotlib.pyplot as plt
from mpl_toolkits import mplot3d
import time as tm

import toolkit

from paneller import *
from utils import *

'''
Script to test Euler solution with a NACA 4412 straight wing
'''

xdisc=20
airfoil, extra, intra=read_afl('n4412', ext_append=True, remove_TE_gap=True, disc=xdisc, extra_intra=True)
b=2.0
c=0.3
Uinf=0.2
ys=np.linspace(b/2, -b/2, 20)
#ys=np.sin(np.linspace(-pi/2, pi/2, 30))*b/2
elip_factor=np.sqrt((b**2/4-0.9*ys**2)*4/b**2)
#elip_factor=np.array([1.0]*len(ys))
totlist=[]
conlist=[]
n=0
for i in range(np.size(airfoil, 0)):
    totlist+=[[]]
    for j in range(len(ys)):
        totlist[-1]+=[np.array([airfoil[i, 0]*elip_factor[j]*c, ys[j], airfoil[i, 1]*elip_factor[j]*c])]
for i in range(len(ys)-1):
    conlist+=[[i, i+(len(ys)-1)*(np.size(airfoil, 0)-2)]]
sld=Solid(sldlist=[totlist], wraparounds=[[1]])
sld.end_preprocess()
sld.genwakepanels(wakecombs=conlist, a=radians(5.0))
sld.plotgeometry(ylim=[-b/2, b/2], xlim=[-b/2, b/2], zlim=[-b/2, b/2])
t=tm.time()
sld.genvbar(Uinf, a=radians(5.0))
sld.gennvv()
print('Pre-processing and geometry gen: '+str(tm.time()-t))
t=tm.time()
sld.genaicm()
print('AIC matrix generation: '+str(tm.time()-t))
t=tm.time()
sld.solve(damper=0.001)
sld.calcpress(Uinf=Uinf)
print('Solution and post-processing: '+str(tm.time()-t))
sld.plotgeometry(ylim=[-b/2, b/2], xlim=[-b/2, b/2], zlim=[-b/2, b/2])
#sld.plotgeometry()

xlocs1=[]
ylocs1=[]
cps1=[]
xlocs2=[]
ylocs2=[]
cps2=[]
for i in range(sld.npanels):
    if sld.panels[i].nvector[2]>0:
        xlocs1+=[sld.panels[i].colpoint[0]]
        ylocs1+=[sld.panels[i].colpoint[1]]
        cps1+=[sld.Cps[i]]
    else:
        xlocs2+=[sld.panels[i].colpoint[0]]
        ylocs2+=[sld.panels[i].colpoint[1]]
        cps2+=[sld.Cps[i]]

fig=plt.figure()
ax=plt.axes(projection='3d')

ax.scatter3D(xlocs1, ylocs1, cps1, 'red')
ax.scatter3D(xlocs2, ylocs2, cps2, 'blue')
plt.show()

xlocs1=[]
ylocs1=[]
g1=[]
xlocs2=[]
ylocs2=[]
g2=[]
for i in range(sld.npanels):
    if sld.panels[i].nvector[2]>0:
        xlocs1+=[sld.panels[i].colpoint[0]]
        ylocs1+=[sld.panels[i].colpoint[1]]
        g1+=[sld.solution[i]]
    else:
        xlocs2+=[sld.panels[i].colpoint[0]]
        ylocs2+=[sld.panels[i].colpoint[1]]
        g2+=[sld.solution[i]]

fig=plt.figure()
ax=plt.axes(projection='3d')

ax.scatter3D(xlocs1, ylocs1, g1, 'red')
ax.scatter3D(xlocs2, ylocs2, g2, 'blue')
plt.show()

sld.plotnormals(ylim=[-b/2, b/2], xlim=[-b/2, b/2], zlim=[-b/2, b/2], factor=0.01)


import numpy as np
import os
from math import *
from LovelacePM import *
import matplotlib.pyplot as plt

ordir=os.getcwd()
os.chdir(os.path.dirname(os.path.abspath(__file__)))

Uinf=200/3.6 #200kph

fuselage_height=2.2
cabin_screen_height=0.8
motor_spinner_height=0.9
motor_width=1.2
radiator_height=0.7
fuselage_width=1.6
motor_length=2.5
cabin_aft_projection=0.5
screen_aft_projection=1.0
cabin_length=2.4
tailcone_length=5.0
tailcone_height=1.1
spinner_length=0.3

wing_height=0.3
b=13.0; c=2.0; taper=0.5; dihedron=5.0

emph_b=4.0
emph_c=1.0
emph_taper=0.6
emph_height=1.1
emph_offset=0.6

empv_b=2.0
empv_c=1.0
empv_taper=0.6
empv_height=1.1
empv_offset=0.6

sld=Solid()

fusdefsect=smooth_angle_defsect_function(r_1x=0.5, r_2x=0.5, r_2y=0.5, r_1y=0.5, ldisc=20, thdisc=10)

sects=[fusdefsect(center=np.array([-spinner_length-motor_length, 0.0, motor_spinner_height]), R=0.0), \
    fusdefsect(center=np.array([-motor_length, 0.0, motor_spinner_height]), R=motor_width/2, z_expand=radiator_height/motor_width), \
        fusdefsect(center=np.array([-screen_aft_projection, 0.0, (fuselage_height-cabin_screen_height)/2]), R=(fuselage_height-cabin_screen_height)/2, \
            y_expand=motor_width/(fuselage_height-cabin_screen_height)), \
            fusdefsect(center=np.array([-cabin_aft_projection, 0.0, fuselage_height/2]), R=fuselage_width/2, z_expand=fuselage_height/fuselage_width), \
                fusdefsect(center=np.array([cabin_length-cabin_aft_projection, 0.0, fuselage_height/2]), R=fuselage_width/2, z_expand=fuselage_height/fuselage_width), \
                    fusdefsect(center=np.array([cabin_length-cabin_aft_projection+tailcone_length, 0.0, tailcone_height]), R=0.0)]

fuselage=body(sld, sections=sects)

root_1=wing_section(afl='atr_foil', afldir=os.getcwd(), c=c, CA_position=np.array([0.0, -fuselage_width/2, wing_height]), xdisc=20)
tip_1=wing_section(afl='atr_foil', afldir=os.getcwd(), c=c*taper, CA_position=np.array([0.0, -b/2, wing_height+tan(radians(dihedron))*b/2]), xdisc=20, closed=True)
wing1quad=wing_quadrant(sld, sect1=tip_1, sect2=root_1)
wing1=wing(sld, wingquads=[wing1quad])
root_2=wing_section(afl='atr_foil', afldir=os.getcwd(), c=c, CA_position=np.array([0.0, fuselage_width/2, wing_height]), xdisc=20)
tip_2=wing_section(afl='atr_foil', afldir=os.getcwd(), c=c*taper, CA_position=np.array([0.0, b/2, wing_height+tan(radians(dihedron))*b/2]), xdisc=20, closed=True)
wing2quad=wing_quadrant(sld, sect1=root_2, sect2=tip_2)
wing2=wing(sld, wingquads=[wing2quad])

wing1.trim_bybody(fuselage, sectside=2)
wing2.trim_bybody(fuselage, sectside=1)

horzemp_tip1=wing_section(afl='atr_emp', afldir=os.getcwd(), CA_position=np.array([cabin_length-cabin_aft_projection+tailcone_length-0.75*emph_c-emph_offset, -emph_b/2, emph_height]), \
    c=emph_c*emph_taper, inverse=True, xdisc=20, closed=True)
horzemp_root1=wing_section(afl='atr_emp', afldir=os.getcwd(), CA_position=np.array([cabin_length-cabin_aft_projection+tailcone_length-0.75*emph_c-emph_offset, 0.0, emph_height]), \
    c=emph_c, inverse=True, xdisc=20)
horzemp1_quad=wing_quadrant(sld, sect1=horzemp_tip1, sect2=horzemp_root1)
horzemp1=wing(sld, wingquads=[horzemp1_quad])

horzemp1.trim_bybody(fuselage, sectside=2)

horzemp_tip2=wing_section(afl='atr_emp', afldir=os.getcwd(), CA_position=np.array([cabin_length-cabin_aft_projection+tailcone_length-0.75*emph_c-emph_offset, emph_b/2, emph_height]), \
    c=emph_c*emph_taper, inverse=True, xdisc=20, closed=True)
horzemp_root2=wing_section(afl='atr_emp', afldir=os.getcwd(), CA_position=np.array([cabin_length-cabin_aft_projection+tailcone_length-0.75*emph_c-emph_offset, 0.0, emph_height]), \
    c=emph_c, inverse=True, xdisc=20)
horzemp2_quad=wing_quadrant(sld, sect2=horzemp_tip2, sect1=horzemp_root2)
horzemp2=wing(sld, wingquads=[horzemp2_quad])

horzemp2.trim_bybody(fuselage, sectside=1)

vertemp_root=wing_section(c=empv_c, afl='atr_rud', afldir=os.getcwd(), CA_position=np.array([cabin_length-cabin_aft_projection+tailcone_length-0.75*empv_c-empv_offset, 0.0, empv_height]), \
    gamma=-90.0, xdisc=15)
vertemp_tip=wing_section(c=empv_c*empv_taper, afl='atr_rud', afldir=os.getcwd(), CA_position=np.array([cabin_length-cabin_aft_projection+tailcone_length-0.75*empv_c-empv_offset, 0.0, empv_height+empv_b]), \
    gamma=-90.0, xdisc=15, closed=True)
vertemp_quad=wing_quadrant(sld, sect1=vertemp_tip, sect2=vertemp_root)
vertemp=wing(sld, wingquads=[vertemp_quad])

vertemp.trim_bybody(fuselage, sectside=2)

S, mac, _=wing1.calc_reference()
acft=aircraft(sld, elems=[fuselage, wing1, wing2, horzemp1, horzemp2, vertemp], cref=mac, Sref=2*S, bref=b, CG=np.array([0.0, 0.0, fuselage_height/2]))

acft.plot_input()

wing1.patchcompose(ydisc=20)
wing2.patchcompose(ydisc=20)
horzemp1.patchcompose(ydisc=10)
horzemp2.patchcompose(ydisc=10)
vertemp.patchcompose(ydisc=10)
fuselage.patchcompose(xdisc=100, thdisc_upright=5, thdisc_downright=5, thdisc_upleft=5, thdisc_downleft=5, leftqueue=[wing1, horzemp1], rightqueue=[wing2, horzemp2], upqueue=[vertemp])

closex=cabin_length-cabin_aft_projection+tailcone_length-0.75*emph_c-emph_offset

#inertial data from https://www.researchgate.net/publication/317371325_Design_Methodology_and_Flight_Test_Protocols_for_a_Dynamically-Scaled_General_Aviation_Aircraft/figures?lo=1&utm_source=google&utm_medium=organic

#conversion to metric units:
m=2260*0.453592
Ixx=14.59390*948*0.3048**2
Iyy=14.59390*1.346*0.3048**2
Izz=14.59390*1.967*0.3048**2
acft.edit_parameters({'Uinf':Uinf})

acft.addwake()
acft.plotgeometry()#(xlim=[closex-0.5, closex+0.5], zlim=[0.6, 1.6], ylim=[-0.5+tailcone_height, 0.5+tailcone_height])
#sld.plotnormals(xlim=[closex-0.5, closex+0.5], zlim=[0.6, 1.6], ylim=[-0.5, 0.5], factor=0.05)
acft.eulersolve()
acft.forces_report()
acft.stabreport()
acft.balance(SM=0.1)
acft.addmass(m=m, Ixx=Ixx, Iyy=Iyy, Izz=Izz)
external_history, alpha_history, beta_history, euler_history, time_history=acft.dynamic_simulation(nstep=2.5e3, dt=1e-4, perturbations={'w':10.0})
plt.plot(time_history, alpha_history)
plt.xlabel('t [s]')
plt.ylabel('Var. in angle of attack [rad]')
plt.show()
external_history, alpha_history, beta_history, euler_history, time_history=acft.dynamic_simulation(nstep=2.5e3, dt=1e-4, perturbations={'v':10.0})
plt.plot(time_history, beta_history)
plt.xlabel('t [s]')
plt.ylabel('Var. in sideslip angle [rad]')
plt.show()

os.chdir(ordir)


import numpy as np
import numpy.linalg as lg
import scipy.linalg as slg
import scipy.linalg.lapack as slapack
from math import *
import matplotlib.pyplot as plt
from mpl_toolkits import mplot3d
import time as tm

from paneller import *
from utils import *
from wing import *
from body import *
from aircraft import *
from aerodynamic_output import *

Uinf=1.0
t=tm.time()
c=0.0
thetas=np.linspace(0.0, 2*pi, 100)
ys=np.linspace(-1000.0, 1000.0, 5)
R=1.0
coordlist1=[]

for th in thetas:
    coordlist1+=[[]]
    for i in range(len(ys)):
        coordlist1[-1]+=[np.array([sin(th), ys[i], cos(th)])*R]
        '''
for i in range(len(coordlist1)):
    coordlist1[i]=[np.array([0.0, ys[0], 0.0])]+coordlist1[i]+[np.array([0.0, ys[-1], 0.0])]'''
    

sld=Solid()
horzlines, vertlines, paninds, sldpts=sld.addpatch(coordlist1, wraps=[1])

sld.end_preprocess()

sld.genvbar(Uinf=Uinf)
sld.gennvv()
print('Solid generation and pre-processing: '+str(tm.time()-t))
t=tm.time()
sld.genaicm()
sld.gen_selfinf_mat()
print('Generating AICs: '+str(tm.time()-t))
t=tm.time()
sld.solve(damper=c)
print('Solving and post-processing: '+str(tm.time()-t))

sld.plotgeometry()
plt.scatter([p.colpoint[0] for p in sld.panels], \
    [lg.norm(sld.delphi[i, :]+sld.vbar[i, :])/Uinf \
    for i in range(len(sld.panels))])
plt.xlabel('$x$')
plt.ylabel('$v$')
plt.show()



import numpy as np
import numpy.linalg as lg
import scipy.linalg as slg
import scipy.linalg.lapack as slapack
from math import *
import matplotlib.pyplot as plt
from mpl_toolkits import mplot3d
import time as tm

import toolkit

from paneller import *

'''
Script made to test the Euler solution modules with a flat-plate straight wing
'''

b=1.0
c=1.0

u=0.05

ys=np.linspace(-b/2, b/2, 50)
xs=np.linspace(0.0, c, 50)

wakeon=[]
coords=[]
npan=0
geomlist=[]
for i in range(len(xs)-1):
    coords=[]
    for j in range(len(ys)-1):
        coords+=[np.array([[xs[i+1], xs[i+1], xs[i], xs[i]], \
            [ys[j], ys[j+1], ys[j+1], ys[j]], [0.0, 0.0, 0.0, 0.0]])]
        if i==len(xs)-2:
            wakeon+=[[npan, -1]]
        npan+=1
    geomlist+=[coords]
sld=Solid(sldlist=[geomlist])
sld.genwakepanels(wakecombs=wakeon, a=radians(30.0))
sld.genvbar(u, a=radians(30.0))
sld.gennvv()
t=tm.time()
sld.genaicm()
print(tm.time()-t)
t=tm.time()
sld.solve(damper=0.0)
print(tm.time()-t)
sld.calcpress(Uinf=u)
sld.plotpress()

sld.plotgeometry(wake=False)

sld.calcforces()
print('CFres: '+str(sld.SCFres))
print('CMres: '+str(sld.SCMres))

fig=plt.figure()
ax=plt.axes(projection='3d')

xs=[]
ys=[]
for i in range(sld.npanels):
    xs+=[sld.panels[i].colpoint[0]]
    ys+=[sld.panels[i].colpoint[1]]
ax.scatter3D(xs, ys, sld.Cps)
plt.show()


import os
import sys
import datetime
import pickle
import LoveUpdate
ordir=os.getcwd()
os.chdir(os.path.dirname(os.path.abspath(__file__)))
sys.path.insert(0, os.getcwd())
from utils import *
from wing import *
from body import *
from control import *
from aircraft import *
from xfoil_visc import *
from aerodynamic_output import *
from toolkit import *
tdy=LoveUpdate.getdate()
if not os.path.exists('updatestat.lup'):
    lupfile=open('updatestat.lup', 'wb')
    statusdict={'lastupdate':tdy, 'updateme':True}
    _=LoveUpdate.update('LoveUpdate', statusdict['lastupdate']-1)
    pickle.dump(statusdict, lupfile)
    lupfile.close()
lupfile=open('updatestat.lup', 'rb')
statusdict=pickle.load(lupfile)
lupfile.close()
hasupdated=False
hasupdated_lup=False
code_lup=False
code=False
if statusdict['updateme']:
    code_lup, hasupdated_lup, tdy=LoveUpdate.update('LoveUpdate', statusdict['lastupdate'])
    if not code_lup:
        if hasupdated_lup:
            print('LoveUpdate has autoupdated. If you wish to deactivate the autoupdate feature, import LovelacePM in another python instance and use LovelacePM.updatecancel()')
        code, hasupdated, tdy=LoveUpdate.update('LovelacePM', statusdict['lastupdate'])
        code=code and code_lup
        if not code:
            if hasupdated:
                print('LovelacePM has autoupdated. If you wish to deactivate the autoupdate feature, import LovelacePM in another python instance and use LovelacePM.updatecancel()')
            statusdict['lastupdate']=tdy
            lupfile=open('updatestat.lup', 'wb')
            pickle.dump(statusdict, lupfile)
            lupfile.close()
        else:
            print('WARNING: error in autoupdate. Check internet connection if you wish to update LovelacePM. Moving on')
    else:
        print('WARNING: error in autoupdate. Check internet connection if you wish to update LoveUpdate. Moving on')
else:
    print('WARNING: LovelacePM autoupdate is deactivated. if you wish to activate autoupdate, import LovelacePM in another python instance and use LovelacePM.updateset()')
def updateset():
    pdir=os.getcwd()
    os.chdir(os.path.dirname(os.path.abspath(__file__)))
    lupfile=open('updatestat.lup', 'rb')
    statusdict=pickle.load(lupfile)
    lupfile.close()
    statusdict['updateme']=True
    lupfile=open('updatestat.lup', 'wb')
    pickle.dump(statusdict, lupfile)
    lupfile.close()
    os.chdir(pdir)
def updatecancel():
    pdir=os.getcwd()
    os.chdir(os.path.dirname(os.path.abspath(__file__)))
    lupfile=open('updatestat.lup', 'rb')
    statusdict=pickle.load(lupfile)
    lupfile.close()
    statusdict['updateme']=False
    lupfile=open('updatestat.lup', 'wb')
    pickle.dump(statusdict, lupfile)
    lupfile.close()
    os.chdir(pdir)
del code, code_lup, lupfile, tdy, hasupdated_lup, hasupdated
os.chdir(ordir)
del ordir



import numpy as np
import numpy.linalg as lg
import scipy.linalg as slg
import scipy.linalg.lapack as slapack
import scipy.sparse.linalg as splg
import scipy.interpolate as sinterp
from math import *
import matplotlib.pyplot as plt
from mpl_toolkits import mplot3d
import scipy.sparse as sps
import time as tm

import toolkit
from utils import *
from body import *
from control import *

class wing_section: #class to define wing section based on airfoil info
    def __init__(self, afldir='', c=1.0, incidence=0.0, gamma=0.0, CA_position=np.array([0.0, 0.0, 0.0]), afl='n4412', \
        header_lines=1, xdisc=10, remove_TE_gap=True, inverse=False, xstrategy=lambda x: (np.sin(pi*x-pi/2)+1)/2, closed=False, \
            correction=None, Re=2e6, sweep=0.0):
        #closed: return camberline as set of points. for closing wings
        #control axpercs: axis percentage of chord position
        self.CA_position=CA_position
        self.c=c
        self.closed=closed
        self.points=wing_afl_positprocess(read_afl(afldir=afldir, afl=afl, ext_append=True, header_lines=header_lines, disc=xdisc, \
            remove_TE_gap=remove_TE_gap, incidence=radians(incidence), inverse=inverse, strategy=xstrategy, sweep=radians(sweep)), \
                gamma=radians(gamma), c=c, xpos=CA_position[0], ypos=CA_position[1], zpos=CA_position[2])
        self.inverted=inverse
        self.correction=correction
        if self.hascorrection():
            self.getcorrection(Re=Re)
    def hascorrection(self): #check whether or not external polar is fed to post-processor
        return self.correction!=None
    def getcorrection(self, Re=2e6):
        self.alphas, self.Cls, self.Cds, self.Cms=self.correction(Re=Re, inverse=self.inverted)
    def addcontrols(self, controls=[], control_multipliers=[], control_axpercs=[]):
        if not self.hascontrol():
            self.controls=controls
            inlen=0
        else:
            inlen=len(self.controls)
            self.controls+=controls
        if len(control_multipliers)==0:
            control_multipliers=[1.0]
        trimlist(len(controls), control_multipliers)
        try:
            trimlist(len(controls), control_axpercs)
        except:
            control_axpercs=[0.75]
            print('WARNING: control axis position absent. Using 0.75 instead.')
            trimlist(len(controls), control_axpercs)
        self.control_multipliers=control_multipliers
        #list to contain the indexes of points to be rotated by each panel's deflection
        self.control_ptinds=[]
        for i in range(len(self.controls)):
            self.control_ptinds+=[[]]
            inds=np.argwhere(self.points[:, 0]>control_axpercs[i]*self.c+self.CA_position[0]-self.c/4)
            for ptind in [index[0] for index in inds]:
                self.control_ptinds[-1]+=[ptind]
        return list(range(inlen, len(self.controls)))
    def hascontrol(self):
        return hasattr(self, 'controls')
    def hasdeflection(self): #define if there is any deflection in controls in section in question
        return any([cntrl.DOF.state!=0.0 for cntrl in self.controls])
    def getinthick(self, eta=0.5, x=0.0): #get point in x, at thickness percentage eta
        LE_ind=np.argmin(self.points[:, 0])
        extra=self.points[0:LE_ind+1, :]
        intra=self.points[np.arange(np.size(self.points, axis=0)-1, LE_ind-1, -1), :]
        intrapty=np.interp(x, intra[:, 0], intra[:, 1])
        intraptz=np.interp(x, intra[:, 0], intra[:, 2])
        extrapty=np.interp(x, extra[:, 0], extra[:, 1])
        extraptz=np.interp(x, extra[:, 0], extra[:, 2])
        return np.array([x, eta*extrapty+(1.0-eta)*intrapty, eta*extraptz+(1.0-eta)*intraptz])
    def applycontrols(self, ths, control_inds):
        if type(ths)!=list:
            ths=[ths]
        trimlist(len(control_inds), ths)
        points=np.array(list(self.points))
        for i in control_inds:
            for ptind in self.control_ptinds[i]:
                pt=np.reshape(points[ptind, :], (3))
                points[ptind, :]=self.controls[i].axis.control_rot_func(pt, self.controls[i].multiplier*ths[i])
        return points

class wing_quadrant: #wing region between two airfoil sections
    def __init__(self, sld, sect1=None, sect2=None, control_names=[], control_axpercs_x=[], control_axpercs_thickness=[], \
        control_multipliers=[]):
        #by convention, set section 1 as the rightmost section in the quadrant (positive in y axis)
        #vertical fins should have sect1 as their upmost section
        #control_axperc_thickness: list of lists. in sublist, respectively, left and right section percentage of thickness for axis.
        #same with cotrol_axpercs_x
        self.sld=sld
        self.sect1=sect1
        self.sect2=sect2
        self.wakecomb=[]
        self.panstrips_extra=[]
        self.panstrips_intra=[]
        self.acft=None
        if len(control_names)!=0:
            self.controls={}
            if len(control_axpercs_x)==0:
                control_axpercs_x=[[0.75, 0.75]]
                print('WARNING: no control percentage of chord provided for wing quadrant constructor. 0.75 used instead by default.')
            if len(control_axpercs_thickness)==0:
                control_axpercs_thickness=[[0.5, 0.5]]
                print('WARNING: no control percentage of thickness provided for wing quadrant constructor. 0.5 used instead by default.')
            if len(control_multipliers)==0:
                control_multipliers=[1.0]
            trimlist(len(control_names), control_axpercs_x)
            trimlist(len(control_names), control_axpercs_thickness)
            trimlist(len(control_names), control_multipliers)
            for i in range(len(control_names)):
                xax1=np.interp(control_axpercs_x[i][0], np.array([0.0, 1.0]), np.array([self.sect1.CA_position[0]-self.sect1.c/4, \
                    self.sect1.CA_position[0]+3*self.sect1.c/4]))
                pax1=self.sect1.getinthick(x=xax1, eta=control_axpercs_thickness[i][0])
                xax2=np.interp(control_axpercs_x[i][1], np.array([0.0, 1.0]), np.array([self.sect2.CA_position[0]-self.sect2.c/4, \
                    self.sect2.CA_position[0]+3*self.sect2.c/4]))
                pax2=self.sect2.getinthick(x=xax2, eta=control_axpercs_thickness[i][1])
                self.controls[control_names[i]]=control(p0=pax1, p1=pax2, multiplier=control_multipliers[i])
            self.sect1_control_indlist=self.sect1.addcontrols(controls=[self.controls[k] for k in self.controls], \
                control_multipliers=[control_multipliers[i] for i in range(len(control_names))], \
                    control_axpercs=[control_axpercs_x[i][0] for i in range(len(control_names))])
            self.sect2_control_indlist=self.sect2.addcontrols(controls=[self.controls[k] for k in self.controls], \
                control_multipliers=[control_multipliers[i] for i in range(len(control_names))], \
                    control_axpercs=[control_axpercs_x[i][1] for i in range(len(control_names))])
    def set_aircraft(self, acft):
        self.acft=acft #define aircraft structure related to instance
    def trim_bybody(self, contactbody, sectside=2, tolerance=0.00005):
        #trim wing section by body contact, for abutment
        if sectside==2:
            ps=self.sect1.points
            us=self.sect2.points-self.sect1.points
            for i in range(np.size(self.sect2.points, 0)):
                newpt, errorcode=contactbody.find_body_intersect(ps[i, :], us[i, :], tolerance=tolerance)
                if errorcode:
                    print('An error has been detected while trimming a wing section to a fuselage. Please check geometry to verify whether all requested intersections are possible,', \
                        'or change requested discretization for fuselage input.')
                else:
                    self.sect2.points[i, :]=newpt
        else:
            ps=self.sect2.points
            us=self.sect1.points-self.sect2.points
            for i in range(np.size(self.sect2.points, 0)):
                newpt, errorcode=contactbody.find_body_intersect(ps[i, :], us[i, :], tolerance=tolerance)
                if errorcode:
                    print('An error has been detected while trimming a wing section to a fuselage. Please check geometry to verify whether all requested intersections are possible,', \
                        'or change requested discretization for fuselage input.')
                else:
                    self.sect1.points[i, :]=newpt
    def plot_input(self, fig=None, ax=None, show=False, xlim=[], \
        ylim=[], zlim=[], colour='blue'): #plot input geometry data
        if fig==None:
            fig=plt.figure()
        if ax==None:
            plt.axes(projection='3d')
        
        ax.plot3D(self.sect1.points[:, 0], self.sect1.points[:, 1], self.sect1.points[:, 2], colour)
        ax.plot3D(self.sect2.points[:, 0], self.sect2.points[:, 1], self.sect2.points[:, 2], colour)
        
        for i in range(np.size(self.sect1.points, 0)):
            ax.plot3D([self.sect1.points[i, 0], self.sect2.points[i, 0]], \
                [self.sect1.points[i, 1], self.sect2.points[i, 1]], \
                    [self.sect1.points[i, 2], self.sect2.points[i, 2]], colour)
        
        if len(xlim)!=0:
            ax.set_xlim3d(xlim[0], xlim[1])
        if len(ylim)!=0:
            ax.set_ylim3d(ylim[0], ylim[1])
        if len(zlim)!=0:
            ax.set_zlim3d(zlim[0], zlim[1])
        if show:
            plt.show()
    def hascontrol(self): #call to wing_section.hascontrol()
        return hasattr(self, 'controls')
    def patchcompose(self, prevlines={}, strategy=lambda x: (np.sin(pi*x-pi/2)+1)/2, ldisc=20, tolerance=5e-5):
        #controlset:
        #variable including all control DOFs in aircraft object
        #code for prevlines: 'intra_left', 'intra_right', 'extra_...
        #and wing quadrant patch additions
        if self.acft.hascontrol():
            controlset=self.acft.controlset
        lspacing=strategy(np.linspace(0.0, 1.0, ldisc))
        extrasld=[]
        xdisc=int(np.size(self.sect1.points, 0)/2)+1

        #detecting closed extremities
        closedl=self.sect1.closed
        closedr=self.sect2.closed

        #handling inverted airfoil sections
        if self.sect1.inverted:
            sect1_extraorder=range(np.size(self.sect1.points, 0)-1, xdisc-2, -1)
            sect1_intraorder=range(xdisc)
        else:
            sect1_intraorder=range(np.size(self.sect1.points, 0)-1, xdisc-2, -1)
            sect1_extraorder=range(xdisc)
        if self.sect2.inverted:
            sect2_extraorder=range(np.size(self.sect2.points, 0)-1, xdisc-2, -1)
            sect2_intraorder=range(xdisc)
        else:
            sect2_intraorder=range(np.size(self.sect2.points, 0)-1, xdisc-2, -1)
            sect2_extraorder=range(xdisc)

        if self.hascontrol(): #this ought to be changed to patch for controlled sections
            ths=[]
            for name in self.controls:
                ths+=[controlset[name].state]
            s1points=self.sect1.applycontrols(ths, self.sect1_control_indlist)
            s2points=self.sect2.applycontrols(ths, self.sect2_control_indlist)
        else:
            s1points=self.sect1.points
            s2points=self.sect2.points
        
        for i in range(xdisc):
            extrasld+=[[]]
            for eta in lspacing:
                extrasld[-1]+=[eta*s1points[sect1_extraorder[i], :]+(1.0-eta)*s2points[sect2_extraorder[i], :]]
        intrasld=[]
        for i in range(xdisc):
            intrasld+=[[]]
            for eta in lspacing:
                intrasld[-1]+=[eta*s2points[sect2_intraorder[i], :]+(1.0-eta)*s1points[sect1_intraorder[i], :]]
        
        #wake info
        self.wakecombs=[]

        #update prevlines to not include points belonging to controls
        controlpts1=[]
        if self.hascontrol():
            for i in self.sect1_control_indlist:
                controlpts1+=self.sect1.control_ptinds[i]
        controlpts1=set(controlpts1)
        controlpts2=[]
        if self.hascontrol():
            for i in self.sect2_control_indlist:
                controlpts2+=self.sect2.control_ptinds[i]
        controlpts2=set(controlpts2)
        sect1_xdisc=np.size(self.sect1.points, axis=0)
        sect2_xdisc=np.size(self.sect2.points, axis=0)
        
        if self.hascontrol():
            if self.sect1.hasdeflection():
                if 'extra_left' in prevlines:
                    for l in range(len(prevlines['extra_left'])):
                        if l in controlpts1 or l-1 in controlpts1:
                            prevlines['extra_left'][l]=-2
                if 'intra_left' in prevlines:
                    for l in range(len(prevlines['intra_left'])):
                        if sect1_xdisc-1-l in controlpts1 or sect1_xdisc-2-l in controlpts1:
                            prevlines['intra_left'][l]=-2
            if self.sect2.hasdeflection():
                if 'extra_right' in prevlines:
                    for l in range(len(prevlines['extra_right'])):
                        if l in controlpts2 or l-1 in controlpts2:
                            prevlines['extra_right'][l]=-2
                if 'intra_right' in prevlines:
                    for l in range(len(prevlines['intra_right'])):
                        if sect2_xdisc-1-l in controlpts2 or sect2_xdisc-2-l in controlpts2:
                            prevlines['intra_right'][l]=-2
            
        prev={}
        if 'extra_right' in prevlines:
            prev['right']=prevlines['extra_right']
        if 'extra_left' in prevlines:
            prev['left']=prevlines['extra_left']
        horzlines, vertlines, paninds, points=self.sld.addpatch(extrasld, prevlines=prev, tolerance=tolerance)
        self.panstrips_extra=[[paninds[i][j] for i in range(len(paninds))] for j in range(len(paninds[0]))]
        self.paninds=[]
        for panlist in paninds:
            self.paninds+=panlist
        extra_paninds=paninds
        TE_extra=paninds[0] 
        TE_extra.reverse() #for wake generation
        self.extraright_lines=[vertlines[i][0] for i in range(len(vertlines))]
        self.extraright_points=[points[i][0] for i in range(len(points))]
        self.extraleft_lines=[vertlines[i][-1] for i in range(len(vertlines))]
        self.extraleft_points=[points[i][-1] for i in range(len(points))]
        LE_lines=[horzlines[-1][i] for i in range(len(horzlines[-1]))]
        LE_lines.reverse()
        TE_lines=[horzlines[0][i] for i in range(len(horzlines[0]))]
        TE_lines.reverse()

        prev={'up':LE_lines, 'low':TE_lines}
        if 'intra_right' in prevlines:
            prev['left']=prevlines['intra_right']
        if 'intra_left' in prevlines:
            prev['right']=prevlines['intra_left']
        horzlines, vertlines, paninds, points=self.sld.addpatch(intrasld, prevlines=prev, invlats=['up', 'low'], tolerance=tolerance)
        self.panstrips_intra=[[paninds[i][j] for i in range(len(paninds))] for j in range(len(paninds[0])-1, -1, -1)]
        for panlist in paninds:
            self.paninds+=panlist
        intra_paninds=paninds
        TE_intra=paninds[0] #for wake generation
        self.intraright_lines=[vertlines[i][-1] for i in range(len(vertlines))]
        self.intraright_points=[points[i][-1] for i in range(len(points))]
        self.intraleft_lines=[vertlines[i][0] for i in range(len(vertlines))]
        self.intraleft_points=[points[i][0] for i in range(len(points))]

        #clean points and lines list from control region
        if self.hascontrol():
            for i in range(len(self.extraleft_points)):
                if i in controlpts1:
                    self.extraleft_points[i]=self.sect1.points[i, :]
            for i in range(len(self.intraleft_points)):
                if sect1_xdisc-i-1 in controlpts1:
                    self.intraleft_points[i]=self.sect1.points[sect1_xdisc-i-1, :]
            for i in range(len(self.extraright_points)):
                if i in controlpts2:
                    self.extraright_points[i]=self.sect2.points[i, :]
            for i in range(len(self.intraright_points)):
                if sect2_xdisc-i-1 in controlpts2:
                    self.intraright_points[i]=self.sect2.points[sect2_xdisc-i-1, :]
            
            if self.sect1.hasdeflection():
                for l in range(len(self.extraleft_lines)):
                    if l in controlpts1 or l-1 in controlpts1:
                        self.sld.panels[extra_paninds[l][-1]].nocirc_enforce(self.extraleft_lines[l]) #avoid excessive circulation from open control surfaces
                        self.extraleft_lines[l]=-2
                for l in range(len(self.intraleft_lines)):
                    if sect1_xdisc-1-l in controlpts1 or sect1_xdisc-2-l in controlpts2:
                        self.sld.panels[intra_paninds[l][0]].nocirc_enforce(self.intraleft_lines[l]) #avoid excessive circulation from open control surfaces
                        self.intraleft_lines[l]=-2
            if self.sect2.hasdeflection():
                for l in range(len(self.extraright_lines)):
                    if l in controlpts2 or l-1 in controlpts2:
                        self.sld.panels[extra_paninds[l][0]].nocirc_enforce(self.extraright_lines[l]) #avoid excessive circulation from open control surfaces
                        self.extraright_lines[l]=-2
                for l in range(len(self.intraright_lines)):
                    if sect2_xdisc-1-l in controlpts2 or sect2_xdisc-2-l in controlpts2:
                        self.sld.panels[intra_paninds[l][-1]].nocirc_enforce(self.intraright_lines[l]) #avoid excessive circulation from open control surfaces
                        self.intraright_lines[l]=-2

        #wake info
        for i in range(len(TE_extra)):
            self.wakecombs+=[[TE_extra[i], TE_intra[i]]]
        
        #closing tips
        if closedl:
            self.close_tip(sectside=1)
        if closedr:
            self.close_tip(sectside=2)
    def calc_reference(self, axis=1): #input for wing's calc reference function
        ys=np.array([self.sect1.CA_position[axis], self.sect2.CA_position[axis]])
        cs=np.array([self.sect1.c, self.sect2.c])
        yspacing=np.linspace(ys[0], ys[1], 100)
        cspacing=np.interp(yspacing, ys, cs)
        #returns S and parcel for MAC integration
        return abs(np.trapz(cs, x=ys)), abs(np.trapz(cspacing**2, x=yspacing))
    def close_tip(self, sectside=2):
        if sectside==2:
            extra_paninds=self.panstrips_extra[0]
            for i in range(len(extra_paninds)):
                if self.extraright_lines[i]!=-2:
                    self.sld.panels[extra_paninds[i]].nocirc_enforce(self.extraright_lines[i])
            intra_paninds=self.panstrips_intra[0]
            for i in range(len(intra_paninds)):
                if self.intraright_lines[i]!=-2:
                    self.sld.panels[intra_paninds[i]].nocirc_enforce(self.intraright_lines[i])
        else:
            extra_paninds=self.panstrips_extra[-1]
            for i in range(len(extra_paninds)):
                if self.extraleft_lines[i]!=-2:
                    self.sld.panels[extra_paninds[i]].nocirc_enforce(self.extraleft_lines[i])
            intra_paninds=self.panstrips_intra[-1]
            for i in range(len(intra_paninds)):
                if self.intraleft_lines[i]!=-2:
                    self.sld.panels[intra_paninds[i]].nocirc_enforce(self.intraleft_lines[i])
    def deriv_incmat(self): #returns lambda function to calculate the derivative of a rotation matrix around the wing quadrant's axis. Used for design variable gradients
        v=np.array([0.0, self.sect2.CA_position[1]-self.sect1.CA_position[1], self.sect2.CA_position[2]-self.sect1.CA_position[2]])
        v/=lg.norm(v)
        u=np.array([1.0, 0.0, 0.0])
        u-=(u@v)*v
        u/=lg.norm(u)
        w=np.cross(u, v)
        Mtosys=np.vstack((u, v, w))
        return Mtosys.T@(np.array([[0.0, 0.0, 1.0], \
            [0.0, 0.0, 0.0], \
                [-1.0, 0.0, 0.0]]))@Mtosys
    def design_derivs(self, sectside=1): #calculate derivatives for posisions and normal vectors by the incidence of each section (identified in sectside kwarg)
        panrs=np.zeros((len(self.paninds), 3))
        panns=np.zeros((len(self.paninds), 3))
        dists=np.zeros(len(self.paninds))
        wingvec=self.sect2.CA_position-self.sect1.CA_position
        wingvec[0]=0.0
        for i in range(len(self.paninds)):
            panrs[i, :]=self.sld.panels[self.paninds[i]].colpoint
            panns[i, :]=self.sld.panels[self.paninds[i]].nvector
            #calculating distance parallel to the wing
            dists[i]=(panrs[i, :]-self.sect1.CA_position)@wingvec
        dist=wingvec@(self.sect2.CA_position-self.sect1.CA_position)
        axpositx=lambda d: self.sect1.CA_position[0]+d*(self.sect2.CA_position[0]-self.sect1.CA_position[0])/dist
        axposity=lambda d: self.sect1.CA_position[1]+d*(self.sect2.CA_position[1]-self.sect1.CA_position[1])/dist
        axpositz=lambda d: self.sect1.CA_position[2]+d*(self.sect2.CA_position[2]-self.sect1.CA_position[2])/dist
        if sectside==1:
            defs=1.0-dists/dist
        else:
            defs=dists/dist
        panrs[:, 0]-=axpositx(dists)
        panrs[:, 1]-=axposity(dists)
        panrs[:, 2]-=axpositz(dists)
        #calculating derivatives in respect to local rotation
        dRds=self.deriv_incmat()
        panrs=(dRds@panrs.T).T
        panns=(dRds@panns.T).T
        #multiplying by deflection for each y position
        panrs[:, 0]=defs*panrs[:, 0]
        panrs[:, 1]=defs*panrs[:, 1]
        panrs[:, 2]=defs*panrs[:, 2]
        panns[:, 0]=defs*panns[:, 0]
        panns[:, 1]=defs*panns[:, 1]
        panns[:, 2]=defs*panns[:, 2]
        return panrs, panns #returning dr/dksi, dn/dksi
    def calc_coefs(self, alpha=0.0, axis=1): #calculate local sectional coefficients
        if axis==1: #calculate unitary, streamwise direction vectors
            u=np.array([cos(alpha), 0.0, sin(alpha)])
            v=np.array([-sin(alpha), 0.0, cos(alpha)])
            ax=np.array([0.0, 1.0, 0.0])
            perpaxis=2
        elif axis==2:
            u=np.array([cos(alpha), sin(alpha), 0.0])
            v=np.array([-sin(alpha), cos(alpha), 0.0])
            ax=np.array([0.0, 0.0, 1.0])
            perpaxis=1
        else:
            print('WARNING: error in axis provided to wingquad.calc_coefs()')
        self.ys=[]
        self.cs=[]
        CA_posits=[]
        self.Cls=[]
        self.Cms=[]
        self.Cds=[]
        self.Gammas=[]
        for i in range(len(self.panstrips_extra)):
            xposits=np.array([self.sld.panels[pind].colpoint[0] for pind in self.panstrips_extra[i]])
            yposits=np.array([self.sld.panels[pind].colpoint[axis] for pind in self.panstrips_extra[i]])
            aftextreme=min(xposits)
            rearextreme=max(xposits)
            self.cs+=[rearextreme-aftextreme]
            self.ys+=[np.mean(yposits)]
            CA_posits+=[np.array([rearextreme/4+3*aftextreme/4, self.ys[-1], \
                np.interp(self.ys[-1], np.array([self.sect1.CA_position[1], self.sect2.CA_position[1]]), \
                    np.array([self.sect1.CA_position[2], self.sect2.CA_position[2]]))])]
            extracps=np.array([self.sld.Cps[pind] for pind in self.panstrips_extra[i]])
            intracps=np.array([self.sld.Cps[pind] for pind in self.panstrips_intra[i]])
            extrax=xposits-CA_posits[-1][0]
            perposits=np.array([self.sld.panels[pind].colpoint[perpaxis] for pind in self.panstrips_intra[i]])
            perposits-=np.mean(perposits)
            intrax=np.array([self.sld.panels[pind].colpoint[0] for pind in self.panstrips_intra[i]])-CA_posits[-1][0]
            extranu=np.array([self.sld.panels[pind].nvector@u for pind in self.panstrips_extra[i]])
            extranv=np.array([self.sld.panels[pind].nvector@v for pind in self.panstrips_extra[i]])
            intranu=np.array([self.sld.panels[pind].nvector@u for pind in self.panstrips_intra[i]])
            intranv=np.array([self.sld.panels[pind].nvector@v for pind in self.panstrips_intra[i]])
            self.Cls+=[(np.trapz(extracps*extranv, x=extrax)+np.trapz(intracps*intranv, x=intrax))/self.cs[-1]]
            circulation=0.0
            gamma_loc=np.zeros(len(self.panstrips_extra[i]))
            for j in range(len(gamma_loc)):
                vloc=self.sld.delphi[self.panstrips_extra[i][j], :]+self.sld.vbar[self.panstrips_extra[i][j], :]
                vloc-=ax*(vloc@ax)
                gamma_loc[j]=lg.norm(vloc)
            circulation-=np.trapz(gamma_loc*extranv, x=extrax)
            for j in range(len(gamma_loc)):
                vloc=self.sld.delphi[self.panstrips_intra[i][j], :]+self.sld.vbar[self.panstrips_intra[i][j], :]
                vloc-=ax*(vloc@ax)
                gamma_loc[j]=lg.norm(vloc)
            circulation-=np.trapz(gamma_loc*intranv, x=intrax)
            self.Gammas+=[circulation]
            self.Cds+=[(np.trapz(extracps*extranu, x=extrax)+np.trapz(intracps*intranu, x=intrax))/self.cs[-1]]
            self.Cms+=[-(np.trapz(extracps*extranv*extrax, x=extrax)+np.trapz(intracps*intranv*intrax, x=intrax))/self.cs[-1]**2]
        self.Cls=np.array(self.Cls)
        self.Cds=np.array(self.Cds)
        self.Cms=np.array(self.Cms)
        self.ys=np.array(self.ys)
        self.cs=np.array(self.cs)
        self.Gammas=np.array(self.Gammas)

        if self.hascorrection(): #apply viscous correction from external polars
            eta=np.interp(self.ys, np.array([self.sect1.CA_position[perpaxis], self.sect2.CA_position[perpaxis]]), np.array([0.0, 1.0]))
            self.Cls_corrected=np.array([self.sect1.Cls(self.Cls[i])*(1.0-eta[i])+eta[i]*self.sect2.Cls(self.Cls[i]) for i in range(len(eta))])
            self.Cds_corrected=np.array([self.sect1.Cds(self.Cls[i])*(1.0-eta[i])+eta[i]*self.sect2.Cds(self.Cls[i]) for i in range(len(eta))])+self.Cds
            self.Cms_corrected=np.array([self.sect1.Cms(self.Cls[i])*(1.0-eta[i])+eta[i]*self.sect2.Cms(self.Cls[i]) for i in range(len(eta))])+self.Cms
        
        self.perpaxis=perpaxis
    def hascorrection(self):
        return self.sect1.hascorrection() and self.sect2.hascorrection()
    def calc_corrected_forces(self): #returns Ceta (for generic parameter eta) considering external polar corrections in coefficients
        parvec=self.sect2.CA_position-self.sect1.CA_position
        parvec[0]=0.0
        parvec/=lg.norm(parvec)
        xposit=np.interp(self.ys, np.array([self.sect1.CA_position[self.perpaxis], self.sect2.CA_position[self.perpaxis]]), \
            np.array([self.sect1.CA_position[0], self.sect2.CA_position[0]]))
        yposit=np.interp(self.ys, np.array([self.sect1.CA_position[self.perpaxis], self.sect2.CA_position[self.perpaxis]]), \
            np.array([self.sect1.CA_position[1], self.sect2.CA_position[1]]))
        zposit=np.interp(self.ys, np.array([self.sect1.CA_position[self.perpaxis], self.sect2.CA_position[self.perpaxis]]), \
            np.array([self.sect1.CA_position[2], self.sect2.CA_position[2]]))
        dCX=0.0; dCY=0.0; dCZ=0.0; dCl=0.0; dCm=0.0; dCn=0.0
        dmom=np.zeros((len(self.ys), 3))
        dforc=np.zeros((len(self.ys), 3))
        a=self.acft.a
        b=self.acft.b
        p=self.acft.p*self.acft.bref*2
        q=self.acft.q*self.acft.cref*2
        r=self.acft.r*self.acft.bref*2
        for i in range(len(self.ys)):
            posit=np.array([xposit[i], yposit[i], zposit[i]])
            freestream=np.array([cos(a)*cos(b), -cos(a)*sin(b), sin(a)], dtype='double')+np.cross(np.array([-p, q, -r], dtype='double'), posit)
            freestream/=lg.norm(freestream)
            perpvec=self.sect2.CA_position-self.sect1.CA_position
            perpvec[0]=0.0
            perpvec=np.cross(freestream, perpvec)
            perpvec/=lg.norm(perpvec)
            dforc[i, :]=perpvec*self.cs[i]*(self.Cls_corrected[i]-self.Cls[i])+freestream*self.cs[i]*(self.Cds_corrected[i]-self.Cds[i])
            dmom[i, :]=np.cross(posit, dforc[i, :])+(self.cs[i]**2)*(self.Cms_corrected[i]-self.Cms[i])*parvec
        dCX=-np.trapz(dforc[:, 0], x=self.ys)/self.acft.Sref
        dCY=-np.trapz(dforc[:, 1], x=self.ys)/self.acft.Sref
        dCZ=-np.trapz(dforc[:, 2], x=self.ys)/self.acft.Sref
        dCl=-np.trapz(dmom[:, 0], x=self.ys)/(self.acft.Sref*self.acft.bref)
        dCm=-np.trapz(dmom[:, 1], x=self.ys)/(self.acft.Sref*self.acft.cref)
        dCn=-np.trapz(dmom[:, 2], x=self.ys)/(self.acft.Sref*self.acft.bref)
        return dCX, dCY, dCZ, dCl, dCm, dCn

class wing:
    def set_aircraft(self, acft):
        self.acft=acft #define aircraft structure related to instance
        for wngqd in self.wingquads:
            wngqd.set_aircraft(acft)
    def __init__(self, sld, wingquads=[]):
        self.sld=sld
        self.coefavailable=False
        self.wingquads=wingquads
        self.leftclosed=False
        self.rightclosed=False
        axial_vec=wingquads[0].sect1.CA_position-wingquads[-1].sect2.CA_position
        self.axis=np.argmax(np.abs(axial_vec[1:3]))+1
    def patchcompose(self, ystrategy=lambda x: x, ydisc=20, tolerance=5e-5):
        if type(ydisc)==list: #use as list if list is provided, specifying each of the quadrants
            trimlist(len(self.wingquads), ydisc)
        else: #distribute between quadrants proportionally to distance between quadrant section CAs if integer
            ltemp=[]
            for quad in self.wingquads:
                ltemp+=[lg.norm(quad.sect1.CA_position-quad.sect2.CA_position)]
            ltot=sum(ltemp)
            disctemp=[]
            for i in range(len(self.wingquads)-1):
                disctemp+=[floor(np.interp(ltemp[i]/ltot, np.array([0.0, 1.0]), np.array([0, ydisc])))]
            disctemp+=[ydisc-sum(disctemp)]
            ydisc=disctemp
        
        #generate patches one by one
        
        #first patch
        self.wingquads[0].patchcompose(strategy=ystrategy, ldisc=ydisc[0], tolerance=tolerance)
        #all other patches
        for i in range(1, len(self.wingquads)):
            self.wingquads[i].patchcompose(prevlines={'intra_left':self.wingquads[i-1].intraright_lines, 'extra_left':self.wingquads[i-1].extraright_lines}, \
                strategy=ystrategy, ldisc=ydisc[i], tolerance=tolerance)

        self.extraleft_lines=self.wingquads[0].extraleft_lines
        self.extraright_lines=self.wingquads[-1].extraright_lines
        self.intraleft_lines=self.wingquads[0].intraleft_lines
        self.intraright_lines=self.wingquads[-1].intraright_lines
        self.extraleft_points=self.wingquads[0].extraleft_points
        self.extraright_points=self.wingquads[-1].extraright_points
        self.intraleft_points=self.wingquads[0].intraleft_points
        self.intraright_points=self.wingquads[-1].intraright_points

        self.paninds=[]
        for quad in self.wingquads:
            self.paninds+=quad.paninds
    def trim_bybody(self, contactbody, sectside=2, tolerance=0.00005):
        #trim wing section by body contact, for abutment
        if sectside==2:
            self.wingquads[-1].trim_bybody(contactbody, sectside=2, tolerance=tolerance)
        else:
            self.wingquads[0].trim_bybody(contactbody, sectside=1, tolerance=tolerance)
    def calc_coefs(self, alpha=0.0, axis=1):
        self.ys=np.array([])
        self.cs=np.array([])
        self.Cls=np.array([])
        self.Cds=np.array([])
        self.Cms=np.array([])
        self.Gammas=np.array([])
        if any([wngqd.hascorrection() for wngqd in self.wingquads]):
            self.Cls_corrected=np.array([])
            self.Cds_corrected=np.array([])
            self.Cms_corrected=np.array([])
            self.plotcorrections=True
        else:
            self.plotcorrections=False
        for quad in self.wingquads:
            quad.calc_coefs(alpha=alpha, axis=axis)
            self.ys=np.hstack((self.ys, quad.ys))
            self.Cls=np.hstack((self.Cls, quad.Cls))
            self.Cds=np.hstack((self.Cds, quad.Cds))
            self.Cms=np.hstack((self.Cms, quad.Cms))
            self.cs=np.hstack((self.cs, quad.cs))
            self.Gammas=np.hstack((self.Gammas, quad.Gammas))
            if self.plotcorrections:
                self.Cls_corrected=np.hstack((self.Cls_corrected, quad.Cls_corrected))
                self.Cds_corrected=np.hstack((self.Cds_corrected, quad.Cds_corrected))
                self.Cms_corrected=np.hstack((self.Cms_corrected, quad.Cms_corrected))
        order=np.argsort(self.ys)
        self.ys=self.ys[order]
        self.cs=self.cs[order]
        self.Cls=self.Cls[order]
        self.Cds=self.Cds[order]
        self.Cms=self.Cms[order]
        self.Gammas=self.Gammas[order]
        if self.plotcorrections:
            self.Cls_corrected=self.Cls_corrected[order]
            self.Cds_corrected=self.Cds_corrected[order]
            self.Cms_corrected=self.Cms_corrected[order]
        self.coefavailable=True
    def calc_reference(self): #calculate local MAC, surface and span
        #identify wing axis (1 or 2)
        disty=abs(self.wingquads[0].sect1.CA_position[1]-self.wingquads[-1].sect2.CA_position[1])
        distz=abs(self.wingquads[0].sect1.CA_position[2]-self.wingquads[-1].sect2.CA_position[2])
        if distz>disty:
            self.axis=2
            b=distz
        else:
            self.axis=1
            b=disty

        S=0.0
        mac=0.0
        for quad in self.wingquads:
            Sq, macq=quad.calc_reference(axis=self.axis)
            S+=Sq
            mac+=macq
        mac/=S
        return S, mac, b
    def close_tip(self, sectside=2):
        if sectside==2:
            self.wingquads[-1].close_tip(sectside=2)
            self.rightclosed=True
        else:
            self.wingquads[0].close_tip(sectside=1)
            self.leftclosed=True
    def plot_input(self, fig=None, ax=None, show=False, xlim=[], \
        ylim=[], zlim=[], colour='blue'): #plot input geometry data
        if fig==None:
            fig=plt.figure()
        if ax==None:
            plt.axes(projection='3d')
        
        for quad in self.wingquads:
            quad.plot_input(fig=fig, ax=ax, show=False)
        
        if len(xlim)!=0:
            ax.set_xlim3d(xlim[0], xlim[1])
        if len(ylim)!=0:
            ax.set_ylim3d(ylim[0], ylim[1])
        if len(zlim)!=0:
            ax.set_zlim3d(zlim[0], zlim[1])
        if show:
            plt.show()
    def genwakepanels(self, offset=1000.0, a=0.0, b=0.0, wakedisc=10, strategy=lambda x: ((np.exp(x)-1.0)/(exp(1)-1.0))**2):
        prevleft=[]
        for quad in self.wingquads:
            _, prevleft=quad.sld.genwakepanels(wakecombs=quad.wakecombs, wakeinds=[[0, 0]], a=a, b=b, offset=offset, disc=wakedisc, strategy=strategy, prevleft=prevleft, leftinvert=True)
    def hascorrection(self):
        return all([wngqd.hascorrection() for wngqd in self.wingquads])
    def calc_corrected_forces(self): #compose correction forces from each wing quadrant
        dCX=0.0; dCY=0.0; dCZ=0.0; dCl=0.0; dCm=0.0; dCn=0.0
        for wngqd in self.wingquads:
            dCX_quad, dCY_quad, dCZ_quad, dCl_quad, dCm_quad, dCn_quad=wngqd.calc_corrected_forces()
            dCX+=dCX_quad
            dCY+=dCY_quad
            dCZ+=dCZ_quad
            dCl+=dCl_quad
            dCm+=dCm_quad
            dCn+=dCn_quad
        return dCX, dCY, dCZ, dCl, dCm, dCn
    def calc_design_dFM(self, section=0, Uinf=1.0, rho=1.225, CG=np.zeros(3)): #section: index between 0 and len(self.wingquads) to indicate the section to derivate coefficients based on
        #returns derivatives of forces and moments
        drdinc=np.zeros((self.sld.npanels, 3))
        dndinc=np.zeros((self.sld.npanels, 3))
        if section!=0:
            drdinc[np.array(self.wingquads[section-1].paninds), :], dndinc[np.array(self.wingquads[section-1].paninds), :]=self.wingquads[section-1].design_derivs(sectside=2)
        if section!=len(self.wingquads):
            drdinc[np.array(self.wingquads[section].paninds), :], dndinc[np.array(self.wingquads[section].paninds), :]=self.wingquads[section].design_derivs(sectside=1)
        dnvvdinc=np.zeros(self.sld.npanels) #derivative of normal velocity component
        for i in range(self.sld.npanels):
            dnvvdinc[i]=self.sld.vbar[i, :]@dndinc[i, :]
        dCps=self.sld.calc_derivative_dn(Uinf=Uinf, dndksi=dnvvdinc)
        dFs=np.zeros((self.sld.npanels, 3))
        dMs=np.zeros((self.sld.npanels, 3))
        for i in range(self.sld.npanels):
            dFs[i, :]=-rho*Uinf**2*(dCps[i]*self.sld.panels[i].nvector+self.sld.Cps[i]*dndinc[i, :])*self.sld.panels[i].S/2
            dMs[i, :]=np.cross(drdinc[i, :], self.sld.forces[i])+np.cross(self.sld.panels[i].colpoint-CG, dFs[i, :])
        return np.sum(dFs, axis=0), np.sum(dMs, axis=0)


import numpy as np
import numpy.linalg as lg
import scipy.linalg as slg
import scipy.linalg.lapack as slapack
from math import *
import matplotlib.pyplot as plt
from mpl_toolkits import mplot3d
from scipy.interpolate import CubicSpline
import time as tm
import os

import toolkit

def read_afl(afl, afldir='', ext_append=False, header_lines=1, disc=0, strategy=lambda x: (np.sin(pi*x-pi/2)+1)/2, \
    remove_TE_gap=False, extra_intra=False, incidence=0.0, inverse=False, closed=False, sweep=0.0):
    ordir=os.getcwd()
    if len(afldir)==0:
	    os.chdir(ordir)
        #os.chdir(os.path.dirname(os.path.abspath(__file__)))
    else:
	    os.chdir(afldir)
    #read arfoil data points from Selig format file
    if ext_append:
        infile=open(afl+'.dat', 'r')
    else:
        infile=open(afl, 'r')
    aflpts=[]
    alltext=infile.read()
    lines=alltext.split('\n')
    for i in range(header_lines, len(lines)):
        linelist=lines[i].split()
        if len(linelist)!=0:
            aflpts+=[[float(linelist[0]), float(linelist[1])]]
    aflpts=np.array(aflpts)
    if inverse:
        aflpts[:, 1]=-aflpts[:, 1]
    aflpts[:, 1]*=cos(sweep)
    if incidence!=0.0:
        R=np.array([[cos(incidence), sin(incidence)], [-sin(incidence), cos(incidence)]])
        if inverse:
            R=R.T
        aflpts=(R@(aflpts.T)).T/cos(incidence)
    if disc!=0:
        xpts=strategy(np.linspace(1.0, 0.0, disc))
        leading_edge_ind=np.argmin(aflpts[:, 0])
        extra=aflpts[0:leading_edge_ind, :]
        intra=aflpts[leading_edge_ind:np.size(aflpts, 0), :]
        extracs=CubicSpline(np.flip(extra[:, 0]), np.flip(extra[:, 1]))
        extra=np.vstack((xpts, extracs(xpts))).T
        xpts=np.flip(xpts)
        intracs=CubicSpline(intra[:, 0], intra[:, 1])
        intra=np.vstack((xpts, intracs(xpts))).T
        aflpts=np.vstack((extra, intra[1:np.size(intra, 0), :]))
    aflpts[:, 0]-=0.25
    intra[:, 0]-=0.25
    extra[:, 0]-=0.25
    if remove_TE_gap:
        midpoint=(aflpts[-1, :]+aflpts[0, :])/2
        aflpts[-1, :]=midpoint
        aflpts[0, :]=midpoint
    if extra_intra:
        tipind=np.argmin(aflpts[:, 0])
        extra=aflpts[0:tipind, :]
        intra=aflpts[tipind:np.size(aflpts, 0), :]
        extra=np.flip(extra, axis=0)
        return aflpts, extra, intra
    if closed:
        tipind=np.argmin(aflpts[:, 0])
        extra=aflpts[0:tipind+1, :]
        intra=aflpts[tipind:np.size(aflpts, 0), :]
        extra=np.flip(extra, axis=0)
        camberline=(intra+extra)/2
        aflpts[:, 1]=np.interp(aflpts[:, 0], camberline[:, 0], camberline[:, 1])
    os.chdir(ordir)
    return aflpts

def wing_afl_positprocess(afl, gamma=0.0, c=1.0, ypos=0.0, xpos=0.0, zpos=0.0):
    #position airfoil coordinate in 3D axis system
    R=np.array([[1.0, 0.0, 0.0], [0.0, cos(gamma), -sin(gamma)], [0.0, sin(gamma), cos(gamma)]])
    aflnew=(R@np.vstack((afl[:, 0]*c, np.zeros(np.size(afl, 0)), afl[:, 1]*c))).T
    aflnew[:, 1]+=ypos
    aflnew[:, 2]+=zpos
    aflnew[:, 0]+=xpos
    return aflnew

def trimlist(n, l): #trim list to defined length based on first element. For input handling
    if (len(l)<n) and len(l)!=0:
        l+=[l[0]]*(n-len(l))
    return l

def trim_polars(th):
    if th>pi:
        return th-2*pi
    elif th<-pi:
        return th+2*pi

def trim_polars_array(thspacing): #trim polars to eliminate congruous equivalences
    validpos=thspacing>=pi
    thspacing[validpos]-=2*pi
    validpos=thspacing<=-pi
    thspacing[validpos]+=2*pi
    return thspacing

def gen_circdefsect_coords(disc): #present input coordinates for circular defsect
    thetas=np.linspace(-pi, pi, disc)
    return np.vstack((np.sin(thetas), np.cos(thetas))).T

def linear_pts(p1, p2, n, endpoint=False):
    #function to provide linearly interpolated segments in 2D space, serves as tool for other functions in folder
    eta=np.linspace(0.0, 1.0, n, endpoint=endpoint)
    coords=np.zeros((n, 2))
    for i in range(n):
        coords[i, :]=(1.0-eta[i])*p1+eta[i]*p2
    return coords

def elliptic_pts(p1, p2, center, r_x, r_y, th1, th2, n, endpoint=False):
    #function to provide elliptically interpolated segments in 2D space, serves as tool for other functions in folder
    thspacing=np.linspace(th1, th2, n, endpoint=endpoint)
    coords=np.zeros((n, 2))
    coords[:, 0]=np.sin(thspacing)*r_x+center[0]
    coords[:, 1]=np.cos(thspacing)*r_y+center[1]
    return coords

def gen_squaredefsect_coords(disc): #present input coordinates for square defsect
    nside=int(disc/8)
    pts=np.zeros((nside*8, 2))
    #side 1
    pts[0:nside, 0]=np.linspace(0.0, -1.0, nside, endpoint=False)
    pts[0:nside, 1]=-1.0
    #side 2
    pts[nside:3*nside, 1]=np.linspace(-1.0, 1.0, 2*nside, endpoint=False)
    pts[nside:3*nside, 0]=-1.0
    #side 3
    pts[3*nside:5*nside, 0]=np.linspace(-1.0, 1.0, 2*nside, endpoint=False)
    pts[3*nside:5*nside, 1]=1.0
    #side 4
    pts[5*nside:7*nside, 1]=np.linspace(1.0, -1.0, 2*nside, endpoint=False)
    pts[5*nside:7*nside, 0]=1.0
    #side 5
    pts[7*nside:8*nside, 0]=np.linspace(1.0, 0.0, nside, endpoint=True)
    pts[7*nside:8*nside, 1]=-1.0
    
    return pts

def smooth_angle_defsect_coords(r_1x, r_2x, r_1y, r_2y, ldisc=30, thdisc=20):
    #coordinates for body.py's smooth coordinate defsect
    n_low=ldisc
    n_sides=ldisc
    n_up=ldisc
    coords=linear_pts(np.array([0.0, -1.0]), np.array([r_1x-1.0, -1.0]), n_low)
    coords=np.vstack((coords, elliptic_pts(np.array([r_1x-1.0, -1.0]), np.array([-1.0, r_1y-1.0]), np.array([r_1x-1.0, r_1y-1.0]), r_1x, r_1y, -pi, -pi/2, \
        thdisc)))
    coords=np.vstack((coords, linear_pts(np.array([-1.0, r_1y-1.0]), np.array([-1.0, 1.0-r_2y]), n_sides)))
    coords=np.vstack((coords, elliptic_pts(np.array([-1.0, 1.0-r_2y]), np.array([r_2x-1.0, 1.0]), np.array([r_1x-1.0, 1.0-r_2y]), r_2x, r_2y, -pi/2, 0.0, \
        thdisc)))
    coords=np.vstack((coords, linear_pts(np.array([r_2x-1.0, 1.0]), np.array([1.0-r_2x, 1.0]), n_up)))
    coords=np.vstack((coords, elliptic_pts(np.array([1.0-r_2x, 1.0]), np.array([1.0, 1.0-r_2y]), np.array([1.0-r_2x, 1.0-r_2y]), r_2x, r_2y, 0.0, pi/2, \
        thdisc)))
    coords=np.vstack((coords, linear_pts(np.array([1.0, 1.0-r_2y]), np.array([1.0, r_1y-1.0]), n_sides)))
    coords=np.vstack((coords, elliptic_pts(np.array([1.0, r_1y-1.0]), np.array([1.0-r_1x, -1.0]), np.array([1.0-r_1x, r_1y-1.0]), r_1x, r_1y, pi/2, pi, \
        thdisc)))
    coords=np.vstack((coords, linear_pts(np.array([1.0-r_1x, -1.0]), np.array([0.0, -1.0]), n_low, endpoint=True)))
    return coords

def Mtostream(a, b): #define coordinate transformation matrix to convert to streamwise coordinate system
    Mtost=np.zeros((3, 3))
    Mtost[0, :]=np.array([cos(a)*cos(b), -cos(a)*sin(b), sin(a)], dtype='double')
    Mtost[1, :]=np.cross(np.array([0.0, 0.0, 1.0]), Mtost[0, :])
    Mtost[1, :]/=lg.norm(Mtost[1, :])
    Mtost[2, :]=np.cross(Mtost[0, :], Mtost[1, :])
    return Mtost

def Mstreamtouni(a, b): #same but inverted
    return Mtostream(a, b).T

def PG_xmult(beta, a, b): #matrix to which multiply a point array to apply Prandtl-Glauert's correction
    return Mstreamtouni(a, b)@np.diag(np.array([1.0, beta, beta]))@Mtostream(a, b)

def PG_inv_xmult(beta, a, b): #inverse of the matrix before
    return lg.inv(PG_xmult(beta, a, b))

def PG_vtouni(beta, a, b): #returns matrix converting incompressible PG calculated velocities to compressible case
    return Mstreamtouni(a, b)@np.diag(np.array([1.0, beta, beta]))@Mtostream(a, b)


import numpy as np
import numpy.linalg as lg
import scipy.linalg as slg
import scipy.linalg.lapack as slapack
from math import *
import matplotlib.pyplot as plt
from mpl_toolkits import mplot3d
import time as tm

from paneller import *
from utils import *
from wing import *
from body import *
from aircraft import *
from aerodynamic_output import *

a=-2.0
Uinf=0.05
b=5.0
taper=0.3
croot=1.0
quarter_sweep=radians(30)
twist=-3.0
incidence=0.0

sld=Solid()

sect1=wing_section(afl='n65415', CA_position=np.array([b*tan(quarter_sweep)/2, -b/2, 0.0]), c=croot*taper, xdisc=30, incidence=incidence+twist, closed=True)#, xstrategy=lambda x: x)
sect2=wing_section(afl='n65415', CA_position=np.array([0.0, 0.0, 0.0]), c=croot, xdisc=30, incidence=incidence)#, xstrategy=lambda x: x)
sect3=wing_section(afl='n65415', CA_position=np.array([b*tan(quarter_sweep)/2, b/2, 0.0]), c=croot*taper, xdisc=30, incidence=incidence+twist, closed=True)#, xstrategy=lambda x: x)

wng1=wing_quadrant(sld, sect1=sect1, sect2=sect2)
wng2=wing_quadrant(sld, sect1=sect2, sect2=sect3)
wng=wing(sld, wingquads=[wng1, wng2])

acft=aircraft(sld, elems=[wng], Sref=b*croot*(1+taper)/2)
acft.edit_parameters({'a':a, 'Uinf':Uinf})

wng.patchcompose(ydisc=40, ystrategy=lambda x: x)#(np.sin(pi*x-pi/2)+1)/2)

acft.addwake()

'''sld.plotnormals(xlim=[-0.6, 0.6], ylim=[-0.6, 0.6], zlim=[-0.6, 0.6], factor=0.1)
sld.plotnormals(xlim=[-0.2, 0.2], ylim=[-0.8, -0.4], zlim=[-0.2, 0.2], factor=0.1)'''
acft.eulersolve()
acft.forces_report()
acft.stabreport()
plot_Cps(sld, elems=[wng])
plot_Cls(sld, wings=[wng])
plot_Cds(sld, wings=[wng])
plot_Cms(sld, wings=[wng])
plot_gammas(sld, wings=[wng])
sld.plotgeometry(xlim=[-0.6, 0.6], ylim=[-0.6, 0.6], zlim=[-0.6, 0.6])


from LovelacePM import *
import numpy as np

ordir=os.getcwd()
os.chdir(os.path.dirname(os.path.abspath(__file__)))

c=1.0
AR=20.0
b=c*AR

a=10.0

sld=Solid()

sl=wing_section(afl='n0012', c=c, CA_position=np.array([0.0, -b/2, 0.0]), xdisc=30, closed=True)
sr=wing_section(afl='n0012', c=c, CA_position=np.array([0.0, b/2, 0.0]), xdisc=30, closed=True)
wngqd=wing_quadrant(sld, sect1=sl, sect2=sr)
wng=wing(sld, wingquads=[wngqd])
acft=aircraft(sld, elems=[wng])

acft.edit_parameters({'a':a})

wng.patchcompose(ydisc=30)
acft.addwake()

acft.eulersolve()

plot_Cps(sld, elems=[wng])

os.chdir(ordir)